#!/usr/bin/env python
"""
_LoadFromTask_

Oracle implementation of Workflow.LoadFromTask
"""

__all__ = []



from WMCore.WMBS.MySQL.Workflow.LoadFromTask import LoadFromTask as LoadFromTaskMySQL

class LoadFromTask(LoadFromTaskMySQL):
    sql = LoadFromTaskMySQL.sql



#!/usr/bin/env python
"""
_RemoveDeuplicates_

Oracle implementation of Workflow.RemoveDuplicates
"""

from WMCore.WMBS.MySQL.Workflow.RemoveDuplicates import RemoveDuplicates as WorkflowRemoveDuplicatesMySQL


class RemoveDuplicates(WorkflowRemoveDuplicatesMySQL):
    pass



#!/usr/bin/env python
"""
Utilities related to file handling
"""


import io
import os
import glob
import stat
import subprocess
import time
import zlib
import logging

from Utils.Utilities import decodeBytesToUnicode


def findFiles(path, pat):
    """
    Find files within given path and matching given pattern.
    :param path: starting directory path (string)
    :param pat: match pattern (string), e.g. *.py or name of the file
    :return: matched file names
    """
    files = []
    for idir, _, _ in os.walk(path):
        files.extend(glob.glob(os.path.join(idir, pat)))
    return files


def tarMode(tfile, opMode):
    """
    Extract proper mode of operation for given tar file. For instance,
    if op='r' and tfile name is file.tar.gz we should get 'r:gz',
    while if tfile name is file.tar.bz2 we should get 'r':bz2', while
    if tfile name is file.tar we should get 'r', etc.
    :param opMode: mode of operation (string), e.g. 'r', or 'w'
    :param tfile: sandbox tar file name (string)
    :return: mode of operation
    """
    ext = tfile.split(".")[-1]
    if ext == 'tar':
        return opMode
    mode = opMode + ':' + ext
    return mode


def calculateChecksums(filename):
    """
    _calculateChecksums_

    Get the adler32 and crc32 checksums of a file. Return None on error

    Process line by line and adjust for known signed vs. unsigned issues
      http://docs.python.org/library/zlib.html

    The cksum UNIX command line tool implements a CRC32 checksum that is
    different than any of the python algorithms, therefore open cksum
    in a subprocess and feed it the same chunks of data that are used
    to calculate the adler32 checksum.

    """
    adler32Checksum = 1  # adler32 of an empty string
    cksumProcess = subprocess.Popen("cksum", stdin=subprocess.PIPE, stdout=subprocess.PIPE)

    # the lambda basically creates an iterator function with zero
    # arguments that steps through the file in 4096 byte chunks
    with open(filename, 'rb') as f:
        for chunk in iter((lambda: f.read(4096)), b''):
            adler32Checksum = zlib.adler32(chunk, adler32Checksum)
            cksumProcess.stdin.write(chunk)

    cksumProcess.stdin.close()
    cksumProcess.wait()

    cksumStdout = cksumProcess.stdout.read().split()
    cksumProcess.stdout.close()

    # consistency check on the cksum output
    filesize = os.stat(filename)[stat.ST_SIZE]
    if len(cksumStdout) != 2 or int(cksumStdout[1]) != filesize:
        raise RuntimeError("Something went wrong with the cksum calculation !")

    cksumStdout[0] = decodeBytesToUnicode(cksumStdout[0])
    return (format(adler32Checksum & 0xffffffff, '08x'), cksumStdout[0])


def tail(filename, nLines=20):
    """
    _tail_

    A version of tail
    Adapted from code on http://stackoverflow.com/questions/136168/get-last-n-lines-of-a-file-with-python-similar-to-tail
    """
    assert nLines >= 0
    pos, lines = nLines + 1, []

    # make sure only valid utf8 encoded chars will be passed along
    with io.open(filename, 'r', encoding='utf8', errors='ignore') as f:
        while len(lines) <= nLines:
            try:
                f.seek(-pos, 2)
            except IOError:
                f.seek(0)
                break
            finally:
                lines = list(f)
            pos *= 2

    text = "".join(lines[-nLines:])

    return text


def getFileInfo(filename):
    """
    _getFileInfo_

    Return file info in a friendly format
    """

    filestats = os.stat(filename)

    fileInfo = {'Name': filename,
                'Size': filestats[stat.ST_SIZE],
                'LastModification': time.strftime("%m/%d/%Y %I:%M:%S %p", time.localtime(filestats[stat.ST_MTIME])),
                'LastAccess': time.strftime("%m/%d/%Y %I:%M:%S %p", time.localtime(filestats[stat.ST_ATIME]))}
    return fileInfo


def findMagicStr(filename, matchString):
    """
    _findMagicStr_

    Parse a log file looking for a pattern string
    """
    with io.open(filename, 'r', encoding='utf8', errors='ignore') as logfile:
        # TODO: can we avoid reading the whole file
        for line in logfile:
            if matchString in line:
                yield line


def getFullPath(name, envPath="PATH"):
    """
    :param name: file name
    :param envPath: any environment variable specified for path (PATH, PYTHONPATH, etc)
    :return: full path if it is under PATH env
    """
    for path in os.getenv(envPath).split(os.path.pathsep):
        fullPath = os.path.join(path, name)
        if os.path.exists(fullPath):
            return fullPath
    return None


def loadEnvFile(wmaEnvFilePath, logger=None):
    """
    _loadEnvFile_
    A simple function to load an additional bash env file into the current script
    runtime environment
    :param wmaEnvFilePath: The path to the environment file to be loaded
    :return:               True if the script has loaded successfully, False otherwise.
    """
    if not logger:
        logger = logging.getLogger()
    subProc = subprocess.run(['bash', '-c', f'source {wmaEnvFilePath} && python -c "import os; print(repr(os.environ.copy()))" '],
                                   capture_output=True, check=False)
    if subProc.returncode == 0:
        newEnv = eval(subProc.stdout)
        os.environ.update(newEnv)
        if subProc.stderr:
            logger.warning("Environment file: %s loaded with errors:", wmaEnvFilePath)
            logger.warning(subProc.stderr.decode())
        else:
            logger.info("Environment file: %s loaded successfully", wmaEnvFilePath)
        return True
    else:
        logger.error("Failed to load environment file: %s", wmaEnvFilePath)
        logger.error(subProc.stderr.decode())
        return False



#!/usr/bin/env python
"""
_New_

Oracle implementation of Locations.New
"""

from WMCore.WMBS.MySQL.Locations.New import New as NewLocationMySQL


class New(NewLocationMySQL):
    sql = """INSERT INTO wmbs_location (site_name, ce_name, running_slots, pending_slots,
                plugin, cms_name, state, state_time)
                SELECT :location, :cename, :running_slots AS running_slots,
                :pending_slots AS pending_slots,
                :plugin AS plugin, :cmsname AS cms_name,
                (SELECT id FROM wmbs_location_state where name = 'Normal') AS state,
                :state_time AS state_time
                FROM DUAL
              WHERE NOT EXISTS (SELECT site_name FROM wmbs_location
                                WHERE site_name = :location)"""

    pnnSQL = """INSERT /*+ IGNORE_ROW_ON_DUPKEY_INDEX (wmbs_pnns (pnn)) */
                   INTO wmbs_pnns (pnn) VALUES (:pnn)"""

    mapSQL = """INSERT INTO wmbs_location_pnns (location, pnn)
                  SELECT (SELECT id from wmbs_location WHERE site_name = :location),
                         (SELECT id from wmbs_pnns WHERE pnn = :pnn) FROM DUAL
                WHERE NOT EXISTS
                    (SELECT * FROM wmbs_location_pnns WHERE
                      location = (SELECT id from wmbs_location WHERE site_name = :location) AND
                      pnn = (SELECT id from wmbs_pnns WHERE pnn = :pnn))"""



#!/usr/bin/env python
"""
_DeleteWorkflow_

Oracle implementation of DeleteWorkflow

"""
__all__ = []



from WMCore.WMBS.MySQL.Workflow.Delete import Delete as DeleteWorkflowMySQL

class Delete(DeleteWorkflowMySQL):
    sql = DeleteWorkflowMySQL.sql




from WMCore.DataStructs.WMObject import WMObject

import logging

class SplitterFactory(WMObject):
    """
    A splitter factory is called with a subscription. From that subscription it
    returns an instance of the splitting algorithm as a JobFactory. This is then
    called with a job type to return sets of jobs.
    """

    def __init__(self, package='WMCore.JobSplitting'):
        # package is the package the splitter will be loaded from
        self.package = package

    def __call__(self, subscription=None,
                 package='WMCore.DataStructs',
                 generators=[],
                 limit = 0):
        # package is the package output of the splitter will be loaded from
        """
        Instantiate an Subscription.split_algo and
        run that algorithm on subscription
        """
        algorithm = subscription['split_algo']

        module = "%s.%s" % (self.package, algorithm)
        module = __import__(module, globals(), locals(), [algorithm])#, -1)
        splitter = getattr(module, algorithm.split('.')[-1])
        return splitter(package=package,
                        subscription=subscription,
                        generators=generators,
                        limit = limit)



"""
MySQL implementation of WorkUnit.GetByTaskFileLumi which uses the same formatter as GetByID
"""
from __future__ import absolute_import, division, print_function

from WMCore.Database.DBFormatter import DBFormatter
from WMCore.WMBS.MySQL.WorkUnit.GetByID import workUnitFormatter


class GetByTaskFileLumi(DBFormatter):
    """
    MySQL implementation of WorkUnit.GetByTaskFileLumi which uses the workUnitFormatter from GetByID
    """

    sql = ('SELECT wu.id as id, wu.taskid as taskid, wu.retry_count as retry_count,'
           ' wu.last_unit_count as last_unit_count, wu.last_submit_time as last_submit_time, wu.status as status,'
           ' assoc.firstevent as firstevent, assoc.lastevent as lastevent,'
           ' assoc.fileid as fileid, assoc.run as run,  assoc.lumi as lumi'
           ' FROM wmbs_workunit wu'
           ' INNER JOIN wmbs_frl_workunit_assoc assoc ON wu.id = assoc.workunit'
           ' WHERE wu.taskid=:taskid AND assoc.fileid=:fileid AND assoc.run=:run AND assoc.lumi=:lumi')

    def execute(self, taskid=None, fileid=None, run_lumi=None, conn=None, transaction=False):
        """
        Execute the lookup by task, file, run, and lumi
        """

        binds = {'taskid': taskid, 'fileid': fileid, 'run': run_lumi.run, 'lumi': run_lumi.lumis[0]}
        result = self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)
        return self.formatOneDict(result)

    def formatOneDict(self, result):
        """
        _formatOneDict_

        Return the row as a dict
        """

        formattedResult = DBFormatter.formatDict(self, result)[0]
        return workUnitFormatter(formattedResult)



#!/usr/bin/env python
"""
_WorkflowSummary_

List the task name and number of jobs running for a given site and subscription
type.
"""


from WMCore.Database.DBFormatter import DBFormatter
from WMCore.JobStateMachine.Transitions import Transitions

from future.utils import listvalues

class WorkflowSummary(DBFormatter):
    sql = """SELECT MAX(wmbs_workflow.id) AS id, wmbs_workflow.name AS wmspec,
                    COUNT(wmbs_job.id) AS num_job,
                    SUM(wmbs_job.outcome) AS success, wmbs_job_state.name AS state
             FROM wmbs_workflow
               INNER JOIN wmbs_subscription ON
                 wmbs_workflow.id = wmbs_subscription.workflow
               INNER JOIN wmbs_jobgroup ON
                 wmbs_subscription.id = wmbs_jobgroup.subscription
               INNER JOIN wmbs_job ON
                 wmbs_jobgroup.id = wmbs_job.jobgroup
               INNER JOIN wmbs_job_state ON
                 wmbs_job.state = wmbs_job_state.id
            GROUP BY wmbs_workflow.name, wmbs_job_state.name
            ORDER BY id DESC"""

    def failCount(self, result):
        if  result["state"] == 'success' or result["state"] == 'cleanout' \
            or result["state"] == 'exhausted':
            return (result["num_job"] - int(result["success"]))
        return 0

    def pendingCount(self, result):
        if  result["state"] == 'none' or result["state"] == 'new':
            return (result["num_job"] - int(result["success"]))
        return 0

    def processingCount(self, result):

        if  result["state"] != 'success' and result["state"] != 'cleanout' \
            and result["state"] != 'exhausted' and result['state'] != 'none' \
            and result["state"] != 'new':
            return result["num_job"]
        else:
            return 0

    def formatWorkflow(self, results):
        workflow = {}
        tran = Transitions()
        for result in results:
            if result["wmspec"] not in workflow:
                workflow[result["wmspec"]] = {}
                for state in tran.states():
                    workflow[result["wmspec"]][state] = 0

                workflow[result["wmspec"]][result["state"]] = result["num_job"]
                workflow[result["wmspec"]]['total_jobs'] = result["num_job"]
                workflow[result["wmspec"]]["real_success"] = int(result["success"])
                workflow[result["wmspec"]]["id"] = result["id"]
                workflow[result["wmspec"]]["wmspec"] = result["wmspec"]
                workflow[result["wmspec"]]["pending"] = self.pendingCount(result)
                workflow[result["wmspec"]]["real_fail"] = self.failCount(result)
                workflow[result["wmspec"]]['processing'] = self.processingCount(result)
            else:
                workflow[result["wmspec"]][result["state"]] = result["num_job"]
                workflow[result["wmspec"]]['total_jobs'] += result["num_job"]
                workflow[result["wmspec"]]["real_success"] += int(result["success"])
                workflow[result["wmspec"]]["pending"] += self.pendingCount(result)
                workflow[result["wmspec"]]["real_fail"] += self.failCount(result)
                workflow[result["wmspec"]]['processing'] += self.processingCount(result)

        # need to order by id (client side)
        return listvalues(workflow)

    def execute(self, conn = None, transaction = False):
        results = self.dbi.processData(self.sql,
                                       conn = conn, transaction = transaction)
        return self.formatWorkflow(self.formatDict(results))



#!/usr/bin/env python
"""
_JobCountBySubscription_

Monitoring DAO classes for Jobs in WMBS
"""
__all__ = []



from WMCore.WMBS.MySQL.Monitoring.DefaultFormatter import DefaultFormatter

class JobCountBySubscriptionAndRun(DefaultFormatter):
    """
    _JobCountBySubscription_

    return the number of jobs grouped by their status and run for given a subscription (fileset, workflow pair)
    """
    # warning oracle keyword is written upper case for easy replacement - not the best way to handle this
    sql = """SELECT wmbs_file_runlumi_map.run, wmbs_job_state.name AS job_state, count(DISTINCT wmbs_job.id) AS numJobs
             FROM wmbs_job
              INNER JOIN wmbs_job_state ON wmbs_job_state.id = wmbs_job.state
              INNER JOIN wmbs_job_assoc ON wmbs_job_assoc.job = wmbs_job.id
              INNER JOIN wmbs_file_runlumi_map ON wmbs_file_runlumi_map.fileid = wmbs_job_assoc.fileid
              INNER JOIN wmbs_jobgroup ON wmbs_jobgroup.id = wmbs_job.jobgroup
              INNER JOIN wmbs_subscription ON wmbs_subscription.id = wmbs_jobgroup.subscription
              INNER JOIN wmbs_fileset ON wmbs_fileset.id = wmbs_subscription.fileset
              INNER JOIN wmbs_workflow ON wmbs_workflow.id = wmbs_subscription.workflow
             WHERE wmbs_fileset.name = :fileset_name AND wmbs_workflow.name = :workflow_name
                   AND wmbs_file_runlumi_map.run = :run
             GROUP BY wmbs_job_state.name, wmbs_file_runlumi_map.run"""

    def execute(self, fileset_name, workflow_name, run, conn = None, transaction = False):
        """
        _execute_

        Execute the SQL for the given job ID and then format and return
        the result.
        """
        bindVars = {"fileset_name": fileset_name, "workflow_name": workflow_name, "run": run}

        result = self.dbi.processData(self.sql, bindVars, conn = conn,
                                      transaction = transaction)

        return self.formatDict(result)



#!/usr/bin/env python
"""
_NewDataset_

MySQL implementation of DBSBuffer.NewDataset
"""

from WMCore.Database.DBFormatter import DBFormatter

class NewDataset(DBFormatter):
    """
    _NewDataset_

    Check for new dataset existence and then
    insert it
    """

    sql = """INSERT IGNORE INTO dbsbuffer_dataset (path, processing_ver, acquisition_era, valid_status, global_tag, parent, prep_id)
               VALUES (:path, :processing_ver, :acquisition_era, :valid_status, :global_tag, :parent, :prep_id)"""

    def execute(self, datasetPath, acquisitionEra = None, processingVer = 0, validStatus = None,
                globalTag = None, parent = None, prep_id = None, conn = None, transaction = False):
        """
        _execute_

        Insert new dataset
        """
        binds = {"path": datasetPath, 'acquisition_era': acquisitionEra,
                 'processing_ver': processingVer, 'valid_status': validStatus,
                 'global_tag': globalTag, 'parent': parent, 'prep_id': prep_id}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)

        return



from builtins import object


class ConfigDBMapInterface(object):
    """
    Interface for converting the configuration to
    db url string and option dictionary for DBFactory parameters

    the implementation of this class should create db url string.
    and option for DBFactory creation.
    For details of options, reference WMCore.Database.DBFactory
    """

    def __init__(self, config):
        self.config = config

    def getDBUrl(self):
        """
        this should return db string
        i.e. mysql://username@hostname.fnal.gov:3306/TestDB
        """
        raise NotImplementedError("getDBUrl is not implemented")

    def getOption(self):
        """
        this should return options of dict format
        i.e. {'engine_parameters': {'pool_size': 10}}
        """
        raise NotImplementedError("getOption is not implemented")



#!/usr/bin/env python
"""
_Monitoring_

Monitoring DAO classes for Jobs in WMBS
"""
from builtins import map
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter
from functools import reduce

class JobsByState(DBFormatter):
    sql = """select count(wmbs_job.state) as count, wmbs_job_state.name as name from wmbs_job
    join wmbs_job_state on wmbs_job.state=wmbs_job_state.id group by state"""

    def format(self, result):
        list = DBFormatter.formatDict(self, result)
        # [{'count': '2', 'name': 'none'}, {'count': '4', 'name': 'new'}]
        def map_function(item):
            return {item['name'] : item['count']}
        def reduce_function(x, y):
            x.update(y)
            return x
        return reduce(reduce_function, list(map(map_function, list)))



#!/usr/bin/env python
"""
_GetDeletableWorkflows_

MySQL implementation of Workflows.GetDeletableWorkflows

"""

from WMCore.Database.DBFormatter import DBFormatter

class GetDeletableWorkflows(DBFormatter):
    """
    A workflow is finished when all subscriptions associated to it are finished
    and all the other workflows using the same spec are finished as well,
    this DAO searches such workflows. It assumes that a spec is not shared by
    two workflows of the same name.
    """

    # Gets completed workflows.
    # (checks there is no unfinished subscription is left on given workflow)
    # Assumes there is no workflow created without subscription is made. (- need to check)
    # Query: all the workflows - workflows with non finished subscription

    compltetedWFs = """ SELECT name FROM wmbs_workflow
                            WHERE name NOT IN (
                                SELECT DISTINCT ww.name FROM wmbs_workflow ww
                                      INNER JOIN wmbs_subscription ws
                                         ON ws.workflow = ww.id
                                WHERE ws.finished =0)
                      """

    # workflows which has not completed child workflow
    # Child workflow means that the workflow uses output files from the other workflow.
    wfsWithIncompletedChildWFs = """ SELECT DISTINCT ww.name FROM wmbs_workflow ww
                                  INNER JOIN wmbs_subscription ws
                                      ON ws.workflow = ww.id
                                  INNER JOIN wmbs_fileset wfs ON
                                     wfs.id = ws.fileset
                                  INNER JOIN wmbs_fileset_files wfsf ON
                                     wfsf.fileset = wfs.id
                                  INNER JOIN wmbs_file_parent wfp ON
                                     wfp.parent = wfsf.fileid
                                  INNER JOIN wmbs_fileset_files child_fileset ON
                                     child_fileset.fileid = wfp.child
                                  INNER JOIN wmbs_subscription child_subscription ON
                                     child_subscription.fileset = child_fileset.fileset
                                  WHERE child_subscription.finished = 0
                            """

    sql = """SELECT DISTINCT wmbs_workflow.name, wmbs_workflow.spec,
                        wmbs_workflow.id AS workflow_id, wmbs_subscription.id AS sub_id
                 FROM wmbs_subscription
                     INNER JOIN wmbs_workflow ON
                         wmbs_workflow.id = wmbs_subscription.workflow
                     INNER JOIN (%s) complete_workflow ON
                         complete_workflow.name = wmbs_workflow.name
                 WHERE wmbs_workflow.name NOT IN (%s)
              """ % (compltetedWFs, wfsWithIncompletedChildWFs)

    def execute(self, conn = None, transaction = False):
        """
        _execute_

        This DAO is a nested dictionary with the following structure:
        {<workflowName? : {spec : <specURL>,
                           workflows : {<workflowID> : [<subId1>, <subId2>]}
                          }
        }
        """

        #Get the completed workflows and subscriptions
        result = self.dbi.processData(self.sql,
                                      conn = conn, transaction = transaction)
        wfsAndSubs = self.formatDict(result)

        wfs = {}
        for entry in wfsAndSubs:
            if entry['name'] not in wfs:
                wfs[entry['name']] = {'workflows' : {}}
            wfs[entry['name']]['spec'] = entry['spec']
            if entry['workflow_id'] not in wfs[entry['name']]['workflows']:
                wfs[entry['name']]['workflows'][entry['workflow_id']] = []
            wfs[entry['name']]['workflows'][entry['workflow_id']].append(entry['sub_id'])

        return wfs




#!/usr/bin/env python
"""
_Exists_

MySQL implementation of Fileset.Exists
"""
__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Exists(DBFormatter):
    sql = """select id from wmbs_fileset
            where name = :name"""

    def format(self, result):
        result = DBFormatter.format(self, result)
        if len(result) == 0:
            return False
        else:
            return int(result[0][0])

    def getBinds(self, name = None):
        return self.dbi.buildbinds(self.dbi.makelist(name), 'name')

    def execute(self, name = None, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, self.getBinds(name),
                         conn = conn, transaction = transaction)
        return self.format(result)



"""
_GetNumberOfJobsForWorkflowTaskStatus_

MySQL implementation of GetNumberOfJobsForWorkflowTaskStatus

Created on Apr 17, 2013

@author: dballest
"""
__all__ = []

from WMCore.Database.DBFormatter import DBFormatter

class GetNumberOfJobsForWorkflowTaskStatus(DBFormatter):
    """
    Get the number of jobs for a workflow/task combination
    in a given job status.
    """

    sql = """SELECT COUNT(*) FROM wmbs_job
             INNER JOIN wmbs_jobgroup ON wmbs_job.jobgroup = wmbs_jobgroup.id
             INNER JOIN wmbs_subscription ON wmbs_subscription.id = wmbs_jobgroup.subscription
             INNER JOIN wmbs_workflow ON wmbs_subscription.workflow = wmbs_workflow.id
             WHERE wmbs_workflow.name = :workflow AND
                   wmbs_workflow.task = :task AND
                   wmbs_job.state = (SELECT id FROM wmbs_job_state WHERE name = :status)
               """

    def execute(self, workflow, task, status = 'executing', conn = None, transaction = False):
        """
        _execute_

        Execute the SQL and return the number of jobs
        """

        binds = {'workflow': workflow, 'task' : task, 'status' : status}

        result = self.dbi.processData(self.sql, binds,
                                      conn = conn, transaction = transaction)

        return self.format(result)[0][0]



"""
_NewSubscription_

MySQL implementation of DBS3Buffer.NewSubscription

Created on May 2, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter


class NewSubscription(DBFormatter):
    """
    _NewSubscription_

    Create a new subscription in the database
    """

    sql = """INSERT IGNORE INTO dbsbuffer_dataset_subscription
            (dataset_id, site, custodial, priority, subscribed, delete_blocks, dataset_lifetime)
            VALUES (:id, :site, :custodial, :priority, 0, :delete_blocks, :dataset_lifetime)
          """

    def _createPhEDExSubBinds(self, datasetID, subscriptionInfo, custodialFlag):
        """
        Creates the database binds for both custodial and non custodial data
        placements.

        :param datasetID: integer with the dataset id
        :param subscriptionInfo: dictionary object from the request spec
        :param custodialFlag: boolean flag defining whether it's custodial or not
        :return: a list of dictionary binds
        """
        # NOTE: the subscription information is already validated upstream, at
        # the request factory. Thus, there is no need to cast/validate anything
        # at this level.

        # DeleteFromSource is not supported for move subscriptions
        delete_blocks = None
        # if None, force it to 0 as per the table schema
        dataLifetime = subscriptionInfo.get('DatasetLifetime', 0) or 0
        if custodialFlag:
            sites = subscriptionInfo['CustodialSites']
        else:
            sites = subscriptionInfo['NonCustodialSites']
        delete_blocks = 1 if subscriptionInfo.get('DeleteFromSource', False) else None

        binds = []
        for site in sites:
            bind = {'id': datasetID,
                    'site': site,
                    'custodial': 1 if custodialFlag else 0,
                    'priority': subscriptionInfo['Priority'],
                    'delete_blocks': delete_blocks,
                    'dataset_lifetime': dataLifetime}
            binds.append(bind)
        return binds

    def execute(self, datasetID, subscriptionInfo, conn=None, transaction=False):
        """
        _execute_

        Execute the query and retrieve the subscriptions
        """
        binds = self._createPhEDExSubBinds(datasetID, subscriptionInfo, True)
        binds.extend(self._createPhEDExSubBinds(datasetID, subscriptionInfo, False))

        if not binds:
            return

        self.dbi.processData(self.sql, binds=binds, conn=conn, transaction=transaction)



#!/usr/bin/env python
"""
_RESTApi_

A standard class implementing a REST interface. You should configure the
application to point at this class, with a model and formatter class configured:

active.section_('rest')
active.rest.object = 'WMCore.WebTools.RESTApi'
active.rest.templates = os.path.join(WMCore.WMInit.getWMBASE(), '/src/templates/WMCore/WebTools/')
active.rest.database = 'MySQL connection string'
active.rest.section_('model')
active.rest.model.object = 'RESTModel'
active.rest.model.database = 'MySQL connection string'
active.rest.model.templates = os.path.join(WMCore.WMInit.getWMBASE(), '/src/templates/WMCore/WebTools/')
active.rest.section_('formatter')
active.rest.formatter.object = 'RESTFormatter'
active.rest.formatter.templates = '/templates/WMCore/WebTools/'

"""

import traceback

from cherrypy import expose, request, response, HTTPError
from cherrypy.lib.cptools import accept

from WMCore.WebTools.WebAPI import WebAPI
from WMCore.WMFactory import WMFactory
from WMCore.Lexicon import replaceToSantizeURL

class RESTApi(WebAPI):
    """
    Don't subclass this, use the RESTModel as a base for your application code
    """
    __version__ = 1
    def __init__(self, config = {}):

        self._set_formatter(config)

        self._set_model(config)

        self.__doc__ = self.model.__doc__

        WebAPI.__init__(self, config)
        self.methods.update({'handler':{'args':[],
                                 'call':self.model.handler,
                                 'version': 1}})

        # TODO: implement HEAD & TRACE
        self.supporttypes = list(self.formatter.supporttypes)

    def _set_model(self, config):
        """
        Load the model object, and assign it to the RESTAPI
        """
        factory = WMFactory('webtools_factory')
        self.model = factory.loadObject(config.model.object, config)

    def _set_formatter(self, config):
        """
        Load the formatter object, and assign it to the RESTAPI
        """
        factory = WMFactory('webtools_factory')
        self.formatter = factory.loadObject(config.formatter.object, config)

    @expose
    def index(self, **kwargs):
        """
        Return the auto-generated documentation for the API
        """
        return self._buildResponse()

    @expose
    def default(self, *args, **kwargs):
        """
        Return either the documentation or run the appropriate method.
        """
        return self._buildResponse(args, kwargs)

    def _buildResponse(self, args=[], kwargs={}):
        """
        Set the headers for the response appropriately and format the response
        data (e.g. serialise to XML, JSON, RSS/ATOM) or return the documentation
        if no method is specified.
        """
        args = list(args)
        if len(args) == 0 and len(kwargs) == 0:
            self.debug('returning REST documentation')
            types = []
            for m in dir(self.formatter):
                prop = dir(self.formatter.__getattribute__(m))
                if 'exposed' in prop:

                    types.append(m)

            return self.templatepage('RESTAPI', model = self.model,
                                 types = types,
                                 title = self.config.title,
                                 description = self.config.description)
        try:
            data, expires = self.methods['handler']['call'](request.method,
                                                            args, kwargs)
            return self._formatResponse(data, expires,
                                       kwargs.get('return_type', None))

        except HTTPError as h:
            # If something raises an HTTPError assume it's something that should
            # go to the client
            response.status = h.args[0]
            self.debug('call to %s with args: %s kwargs: %s resulted in %s' % (request.method, args, kwargs, h.args[1]))
            return self._formatResponse({'exception': h.args[0],
                                        'type': 'HTTPError',
                                        'message': h.args[1]}, expires=0,
                                        format=kwargs.get('return_type', None))
        except Exception as e:
            # If something raises a generic exception assume the details are private
            # and should not go to the client
            response.status = 500
            debugMsg = replaceToSantizeURL("""call to %s with args: %s kwargs: %s resulted in %s \n
                                              stack trace: %s""" % (request.method, args,
                                                     kwargs, e.__str__(), traceback.format_exc()))
            self.debug(debugMsg)
            return self._formatResponse({'exception': 500,
                                        'type': e.__class__.__name__,
                                        'message': 'Server Error'}, expires=0,
                                        format=kwargs.get('return_type', None))

    def _formatResponse(self, data, expires, format=None):
        """
        data format can be anything API provides, but it will make sense
        to have either dict format or list of dict format.
        """

        acchdr = request.headers.elements('Accept')
        if format:
            datatype = format
        elif len(acchdr) == 1 and '*/*' == str(acchdr[0]):
            datatype = '*/*'
        else:
            datatype = accept(self.supporttypes)

        return self.formatter.format(data, datatype, expires)



#!/usr/bin/env python
"""
_Delete_

MySQL implementation of Subscriptions.Delete

"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Delete(DBFormatter):
    """

    """
    sql = """DELETE FROM wmbs_subscription WHERE id = :id"""

    def getBinds(self, id=None):
        return self.dbi.buildbinds(self.dbi.makelist(id), 'id')

    def format(self, result):
        return True

    def execute(self, id = -1, conn = None, transaction = False):
        self.dbi.processData(self.sql, self.getBinds(id),
                     conn = conn, transaction = transaction)
        return True #or raise



"""
_InsertComponent_

MySQL implementation of InsertWorker
"""

__all__ = []



import time
from WMCore.Database.DBFormatter import DBFormatter

class InsertWorker(DBFormatter):

    sql = """INSERT INTO wm_workers (component_id, name, last_updated, state, pid, poll_interval, cycle_time)
             VALUES ((SELECT id FROM wm_components WHERE name = :component_name),
                     :worker_name, :last_updated, :state, :pid, :poll_interval, :cycle_time)
             """

    def execute(self, componentName, workerName, state, pid, pollInt, cycleTime,
                conn = None, transaction = False):

        binds = {"component_name": componentName,
                 "worker_name": workerName,
                 "last_updated": int(time.time()),
                 "state": state,
                 "pid": pid,
                 "poll_interval": pollInt,
                 "cycle_time": cycleTime}

        self.dbi.processData(self.sql, binds, conn = conn,
                             transaction = transaction)
        return



#!/usr/bin/env python
"""
_ListByStateAndLocation_

MySQL implementation of ListByStateAndLocation

Created on Feb 21, 2013

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class ListByStateAndLocation(DBFormatter):
    """
    DAO to list jobs given certain state and location,
    only returns a list of dictionaries with job information in order
    to match them with BossAir records.
    """
    sql = """SELECT wmbs_job.id, wmbs_job.retry_count,
                    wmbs_job.cache_dir
                FROM wmbs_job
                INNER JOIN wmbs_job_state ON
                    wmbs_job.state = wmbs_job_state.id
                INNER JOIN wmbs_location ON
                    wmbs_location.id = wmbs_job.location
                 WHERE wmbs_job_state.name = :state AND
                     wmbs_location.site_name = :location"""

    def execute(self, state, location, conn = None, transaction = False):
        result = self.dbi.processData(self.sql, {'state' : state, 'location' : location},
                                      conn = conn,
                                      transaction = transaction)
        return self.formatDict(result)



#!/usr/bin/env python
"""
_SetCouchID_

Oracle implementation of Jobs.SetCouchID
"""




from WMCore.WMBS.MySQL.Jobs.SetCouchID import SetCouchID as MySQLSetCouchID

class SetCouchID(MySQLSetCouchID):
    """
    Identical to MySQL version.
    """
    pass



#!/usr/bin/env python
"""
_SRMV2Impl_

Implementation of StageOutImpl interface for SRM Version 2

"""
from builtins import zip, range

import os, re
import logging, tempfile
import subprocess
from subprocess import Popen, PIPE

from WMCore.Storage.StageOutImplV2 import StageOutImplV2
from WMCore.Storage.StageOutError import StageOutError, StageOutFailure

from WMCore.Storage.Execute import runCommandWithOutput as runCommand

_CheckExitCodeOption = True



class SRMV2Impl(StageOutImplV2):
    """
    _SRMV2Impl_

    Implement interface for srmcp v2 command

    """

    run = staticmethod(runCommand)

    def __init__(self):
        StageOutImplV2.__init__(self)
        self.directoryErrorCodes = (1,)


    def createSourceName(self, protocol, pfn):
        """
        _createSourceName_

        SRM uses file:/// urls

        """
        if pfn.startswith('/'):
            return "file:///%s" % pfn
        elif os.path.isfile(pfn):
            return "file:///%s" % os.path.abspath(pfn)
        else:
            return pfn


    def createOutputDirectory(self, targetPFN, stageOut):
        """
        _createOutputDirectory_

        SRMV2 does not create directories,
            see http://sdm.lbl.gov/srm-wg/doc/SRM.v2.2.html#_Toc199734394

        """
        targetdir = os.path.dirname(targetPFN)

        if not stageOut:
            # stage in to local directory - should exist but you never know
            if not os.path.exists(targetdir):
                os.makedirs(targetdir)
            return

        #  // Loop from top level checking existence stop when directory exists
        # // assume first 4 slashes are from srm://host:8443/srm/managerv2?SFN=
        dirs = ["/".join(targetdir.split("/")[0:6+i]) \
                                        for i in range(targetdir.count("/")-4)]
        dirsToCheck = dirs[:]; dirsToCheck.reverse()
        levelToCreateFrom = len(dirs)
        # the -1 in the zip is because we assume that /store is there
        for count, dir in zip(range(len(dirsToCheck), 0, -1), dirsToCheck):
            _, output = self.runCommandWarnOnError(['srmls',\
                                                           '-recursion_depth=0',\
                                                            '-retry_num=%s'\
                                                              % self.numRetries,\
                                                            dir])
            levelToCreateFrom = count # create dirs from here (at least)
            if not output.count('SRM_FAILURE'): # any other codes?
                break

        #  // Create needed directory levels from end of previous loop
        # //  to end of directory structure
        for dir in dirs[levelToCreateFrom - 1:]:
            #print "Create directory: %s" % dir
            self.runCommandWarnOnError(['srmmkdir',
                                        '-retry_num=%s' % self.numRetries,
                                        dir])

    def doDelete(self, pfn, pnn, command, options, protocol  ):
        """
        handle both srm and file pfn types
        """
        if pfn.startswith("srm://"):
            self.runCommandWarnOnNonZero(["srmrm", 'pfn'])
        elif pfn.startswith("file:"):
            self.runCommandWarnOnNonZero(["/bin/rm", "-f", pfn.replace("file://", "", 1)])
        elif pfn.startswith('/'):
            self.runCommandWarnOnNonZero(["/bin/rm", "-f", pfn])
        elif os.path.isfile(pfn):
            self.runCommandWarnOnNonZero(["/bin/rm", "-f", os.path.abspath(pfn)])
        else:
            logging.info("Tried to delete, but nothing knew how")
            logging.info("pfn: %s" % pfn)

    def doTransfer(self, fromPfn, toPfn, stageOut, pnn, command, options, protocol, checksum  ):
        toPfn   = self.createSourceName(protocol, toPfn)
        fromPfn = self.createSourceName(protocol, fromPfn)
        # TODO tee the output to another file
        # attempt copy
        for x in range(self.numRetries):
            (_,reportFile) = tempfile.mkstemp()
            ourCommand = \
                self.generateCommandFromPreAndPostParts(\
                            ['srmcp','-2','-report=%s'%reportFile,
                                          '-retry_num=0'],
                            [fromPfn, toPfn],
                            options)
            self.runCommandWarnOnNonZero(ourCommand)

            if not stageOut:
                remotePFN, localPFN = fromPfn, toPfn.replace("file://", "", 1)
            else:
                remotePFN, localPFN = toPfn, fromPfn.replace("file://", "", 1)


            if _CheckExitCodeOption:
                p1 = Popen(["cat", reportFile], stdout=PIPE)
                p3 = Popen(['cut','-f3','-d',' '], stdin=p1.stdout, stdout=PIPE)
                exitCode = p3.communicate()[0].rstrip()
                logging.info("srmcp exit status: %s" % exitCode)
                p2 = Popen(['grep', '-c', 'SRM_INVALID_PATH',reportFile],stdout=PIPE)
                invalidPathCount = p2.communicate()[0]
                logging.info("got this for SRM_INVALID_PATH: %s" % invalidPathCount)
                if (invalidPathCount and (exitCode == '')):
                    logging.warning("Directory doesn't exist in srmv2 stageout...creating and retrying")
                    self.createOutputDirectory(toPfn,stageOut)
                    continue
                elif ( str(exitCode) != "0" ):
                    logging.error("Couldn't stage out! Error code: %s" % exitCode)
                    self.doDelete(toPfn,None,None,None,None)
                    raise StageOutFailure("srmcp failed! Error code: %s" % exitCode)
                else:
                    logging.info("Tentatively succeeded transfer, will check metadata")
                    break

        localSize = os.path.getsize( localPFN )
        logging.info("Local Size %s" % localSize)

        remotePath = None
        SFN = '?SFN='
        sfn_idx = remotePFN.find(SFN)
        if sfn_idx >= 0:
            remotePath = remotePFN[sfn_idx+5:]
        r = re.compile('srm://([A-Za-z\-\.0-9]*)(:[0-9]*)?(/.*)')
        m = r.match(remotePFN)
        if not m:
            raise StageOutError("Unable to determine path from PFN for " \
                                "target %s." % remotePFN)
        if remotePath == None:
            remotePath = m.groups()[2]
        remoteHost = m.groups()[0]
        #         filesize() { `srm-get-metadata -retry_num=0 %s 2>/dev/null | grep 'size :[0-9]' | cut -f2 -d":"`}
        # the following replaces the above
        logging.info("remote path: %s" % remotePath)
        logging.info("remote host: %s" % remoteHost)
        p1 = Popen(["srmls", '-recursion_depth=0','-retry_num=0', remotePFN], stdout=PIPE)
        p2 = Popen(["grep", remotePath], stdout=PIPE, stdin=p1.stdout)
        p3 = Popen(["grep", '-v', remoteHost], stdout=PIPE, stdin=p2.stdout)
        p4 = Popen(["awk", "{print $1;}"], stdout=PIPE,stdin=p3.stdout)
        remoteSize = p4.communicate()[0]
        logging.info("Localsize: %s Remotesize: %s" % (localSize, remoteSize))
        if int(localSize) != int(remoteSize):
            try:
                self.doDelete(toPfn,None,None,None,None)
            except:
                pass
            raise StageOutFailure("File sizes don't match")

        return toPfn

    def runCommandWarnOnError(self,command):
        return SRMV2Impl.runCommandWarnOnError(self, command)



#!/usr/bin/env python
"""
_NewJobs_

Oracle implementation for creating a new job
"""

from WMCore.BossAir.MySQL.NewJobs import NewJobs as MySQLNewJobs


class NewJobs(MySQLNewJobs):
    """
    _NewJobs_

    Insert new jobs into bl_runjob
    """

    sql = """INSERT INTO bl_runjob (wmbs_id, grid_id, bulk_id, sched_status,
                 retry_count, user_id, location, status_time)
               SELECT :jobid, :gridid, :bulkid,
                 (SELECT id FROM bl_status WHERE name = :status),
                 :retry_count,
                 (SELECT id FROM wmbs_users WHERE cert_dn = :userdn AND group_name = :usergroup AND role_name = :userrole),
                 (SELECT id FROM wmbs_location WHERE site_name = :location),
                 :status_time
               FROM dual
               WHERE NOT EXISTS (SELECT id FROM bl_runjob WHERE wmbs_id = :jobid
                                   AND retry_count = :retry_count)"""



"""
_GetParentInfo_

Oracle implementation of Files.GetParentAndGrandParentInfo
"""




from WMCore.WMBS.MySQL.Files.GetParentInfo import GetParentInfo as GetParentInfoMySQL

class GetParentInfo(GetParentInfoMySQL):
    pass



#!/usr/bin/env python
"""
_TrivialFileCatalog_

Object to contain LFN to PFN mappings from a Trivial File Catalog
and provide functionality to match LFNs against them

Usage:

given a TFC file, invoke readTFC on it. This will return
a TrivialFileCatalog instance that can be used to match LFNs
to PFNs.

Usage: Given a TFC constact string: trivialcatalog_file:/path?protocol=proto


    filename = tfcFilename(tfcContactString)
    protocol = tfcProtocol(tfcContactString)
    tfcInstance = readTFC(filename)

    lfn = "/store/PreProd/unmerged/somelfn.root"

    pfn = tfcInstance.matchLFN(protocol, lfn)


"""

from builtins import next, str, range
from future.utils import viewitems

from future import standard_library
standard_library.install_aliases()

import os
import re

from urllib.parse import urlsplit
from xml.dom.minidom import Document

from WMCore.Algorithms.ParseXMLFile import xmlFileToNode

_TFCArgSplit = re.compile("\?protocol=")


class TrivialFileCatalog(dict):
    """
    _TrivialFileCatalog_

    Object that can map LFNs to PFNs based on contents of a Trivial
    File Catalog
    """

    def __init__(self):
        dict.__init__(self)
        self['lfn-to-pfn'] = []
        self['pfn-to-lfn'] = []
        self.preferredProtocol = None  # attribute for preferred protocol

    def addMapping(self, protocol, match, result,
                   chain=None, mapping_type='lfn-to-pfn'):
        """
        _addMapping_

        Add an lfn to pfn mapping to this instance

        """
        entry = {}
        entry.setdefault("protocol", protocol)
        entry.setdefault("path-match-expr", re.compile(match))
        entry.setdefault("path-match", match)
        entry.setdefault("result", result)
        entry.setdefault("chain", chain)
        self[mapping_type].append(entry)

    def _doMatch(self, protocol, path, style, caller):
        """
        Generalised way of building up the mappings.
        caller is the method from there this method was called, it's used
        for resolving chained rules

        Return None if no match

        """
        for mapping in self[style]:
            if mapping['protocol'] != protocol:
                continue
            if mapping['path-match-expr'].match(path) or mapping["chain"] != None:
                if mapping["chain"] != None:
                    oldpath = path
                    path = caller(mapping["chain"], path)
                    if not path:
                        continue
                splitList = []
                if len(mapping['path-match-expr'].split(path, 1)) > 1:
                    for split in range(len(mapping['path-match-expr'].split(path, 1))):
                        s = mapping['path-match-expr'].split(path, 1)[split]
                        if s:
                            splitList.append(s)
                else:
                    path = oldpath
                    continue
                result = mapping['result']
                for split in range(len(splitList)):
                    result = result.replace("$" + str(split + 1), splitList[split])
                return result

        return None

    def matchLFN(self, protocol, lfn):
        """
        _matchLFN_

        Return the result for the LFN provided if the LFN
        matches the path-match for that protocol

        Return None if no match

        """
        result = self._doMatch(protocol, lfn, "lfn-to-pfn", self.matchLFN)
        return result

    def matchPFN(self, protocol, pfn):
        """
        _matchLFN_

        Return the result for the LFN provided if the LFN
        matches the path-match for that protocol

        Return None if no match

        """
        result = self._doMatch(protocol, pfn, "pfn-to-lfn", self.matchPFN)
        return result

    def getXML(self):
        """
        Converts TFC implementation (dict) into a XML string representation.
        The method reflects this class implementation - dictionary containing
        list of mappings while each mapping (i.e. entry, see addMapping
        method) is a dictionary of key, value pairs.

        """

        def _getElementForMappingEntry(entry, mappingStyle):
            xmlDocTmp = Document()
            element = xmlDocTmp.createElement(mappingStyle)
            for k, v in viewitems(entry):
                # ignore empty, None or compiled regexp items into output
                if not v or (k == "path-match-expr"):
                    continue
                element.setAttribute(k, str(v))
            return element

        xmlDoc = Document()
        root = xmlDoc.createElement("storage-mapping")  # root element name
        for mappingStyle, mappings in viewitems(self):
            for mapping in mappings:
                mapElem = _getElementForMappingEntry(mapping, mappingStyle)
                root.appendChild(mapElem)
        return root.toprettyxml()

    def __str__(self):
        result = ""
        for mapping in ['lfn-to-pfn', 'pfn-to-lfn']:
            for item in self[mapping]:
                result += "\t%s: protocol=%s path-match-re=%s result=%s" % (
                    mapping,
                    item['protocol'],
                    item['path-match-expr'].pattern,
                    item['result'])
                if item['chain'] != None:
                    result += " chain=%s" % item['chain']
                result += "\n"
        return result


def tfcProtocol(contactString):
    """
    _tfcProtocol_

    Given a Trivial File Catalog contact string, extract the
    protocol from it.

    """
    args = urlsplit(contactString)[3]
    value = args.replace("protocol=", '')
    return value


def tfcFilename(contactString):
    """
    _tfcFilename_

    Extract the filename from a TFC contact string.

    """
    value = contactString.replace("trivialcatalog_file:", "")
    value = _TFCArgSplit.split(value)[0]
    path = os.path.normpath(value)
    return path


def readTFC(filename):
    """
    _readTFC_

    Read the file provided and return a TrivialFileCatalog
    instance containing the details found in it

    """
    if not os.path.exists(filename):
        msg = "TrivialFileCatalog not found: %s" % filename
        raise RuntimeError(msg)

    try:
        node = xmlFileToNode(filename)
    except Exception as ex:
        msg = "Error reading TrivialFileCatalog: %s\n" % filename
        msg += str(ex)
        raise RuntimeError(msg)

    parsedResult = nodeReader(node)

    tfcInstance = TrivialFileCatalog()
    for mapping in ['lfn-to-pfn', 'pfn-to-lfn']:
        for entry in parsedResult[mapping]:
            protocol = entry.get("protocol", None)
            match = entry.get("path-match", None)
            result = entry.get("result", None)
            chain = entry.get("chain", None)
            if True in (protocol, match == None):
                continue
            tfcInstance.addMapping(str(protocol), str(match), str(result), chain, mapping)
    return tfcInstance


def loadTFC(contactString):
    """
    _loadTFC_

    Given the contact string for the tfc, parse out the file location
    and the protocol and create a TFC instance

    """
    protocol = tfcProtocol(contactString)
    catalog = tfcFilename(contactString)
    instance = readTFC(catalog)
    instance.preferredProtocol = protocol
    return instance


def coroutine(func):
    """
    _coroutine_

    Decorator method used to prime coroutines

    """

    def start(*args, **kwargs):
        cr = func(*args, **kwargs)
        next(cr)
        return cr

    return start


def nodeReader(node):
    """
    _nodeReader_

    Given a node, see if we can find what we're looking for
    """

    processLfnPfn = {
        'path-match': processPathMatch(),
        'protocol': processProtocol(),
        'result': processResult(),
        'chain': processChain()
    }

    report = {'lfn-to-pfn': [], 'pfn-to-lfn': []}
    processSMT = processSMType(processLfnPfn)
    processor = expandPhEDExNode(processStorageMapping(processSMT))
    processor.send((report, node))
    return report


@coroutine
def expandPhEDExNode(target):
    """
    _expandPhEDExNode_

    If pulling a TFC from the PhEDEx DS, its wrapped in a top level <phedex> node,
    this routine handles that extra node if it exists
    """
    while True:
        report, node = (yield)
        sentPhedex = False
        for subnode in node.children:
            if subnode.name == "phedex":
                target.send((report, subnode))
                sentPhedex = True
        if not sentPhedex:
            target.send((report, node))


@coroutine
def processStorageMapping(target):
    """
    Process everything

    """
    while True:
        report, node = (yield)
        for subnode in node.children:
            if subnode.name == 'storage-mapping':
                target.send((report, subnode))


@coroutine
def processSMType(targets):
    """
    Process the type of storage-mapping

    """
    while True:
        report, node = (yield)
        for subnode in node.children:
            if subnode.name in ['lfn-to-pfn', 'pfn-to-lfn']:
                tmpReport = {'path-match-expr': subnode.name}
                targets['protocol'].send((tmpReport, subnode.attrs.get('protocol', None)))
                targets['path-match'].send((tmpReport, subnode.attrs.get('path-match', None)))
                targets['result'].send((tmpReport, subnode.attrs.get('result', None)))
                targets['chain'].send((tmpReport, subnode.attrs.get('chain', None)))
                report[subnode.name].append(tmpReport)


@coroutine
def processPathMatch():
    """
    Process path-match

    """
    while True:
        report, value = (yield)
        report['path-match'] = value


@coroutine
def processProtocol():
    """
    Process protocol

    """
    while True:
        report, value = (yield)
        report['protocol'] = value


@coroutine
def processResult():
    """
    Process result

    """
    while True:
        report, value = (yield)
        report['result'] = value


@coroutine
def processChain():
    """
    Process chain

    """
    while True:
        report, value = (yield)
        if value == "":
            report['chain'] = None
        else:
            report['chain'] = value



#!/usr/bin/env python
"""
_Load_

MySQL implementation of Masks.Load
"""

__all__ = []



from WMCore.Database.DBFormatter import DBFormatter

class Load(DBFormatter):
    sql = """SELECT DISTINCT FirstEvent, LastEvent, FirstLumi, LastLumi, FirstRun,
             LastRun FROM wmbs_job_mask WHERE job = :jobid"""

    def format(self, results):
        dictList = DBFormatter.formatDict(self, results)

        out = []

        for entry in dictList:
            tmpDict = {}
            tmpDict['FirstEvent'] = entry['firstevent']
            tmpDict['LastEvent']  = entry['lastevent']
            tmpDict['FirstLumi']  = entry['firstlumi']
            tmpDict['LastLumi']   = entry['lastlumi']
            tmpDict['FirstRun']   = entry['firstrun']
            tmpDict['LastRun']    = entry['lastrun']

            out.append(tmpDict)

        return out

    def execute(self, jobid, conn = None, transaction = False):
        binds = self.getBinds(jobid = jobid)
        result = self.dbi.processData(self.sql, binds, conn = conn,
                                      transaction = transaction)
        return self.format(result)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
This script can be used to fetch and update pileup objects through the
MSPileup REST API.

The user provides the MSPileup URL and a JSON file as input, which will be
used to update every single pileup document that it fetches from MSPileup.

Example usage is:
python3 updatePileupObjects.py --url=https://cmsweb.cern.ch --fin=override_data.json

# run script to add new transition record to all MSPileup records if they do not have it
python3 updatePileupObjects.py --add-tran-record -url=https://cmsweb.cern.ch --userDN=<user-dn>
"""
import argparse
import json
import logging
import os
import sys
import urllib.request
import ssl
from http.client import HTTPSConnection

# WMCore services
from Utils.Timers import gmtimeSeconds


class OptionParser():
    """Class to parse the command line arguments"""

    def __init__(self):
        "User based option parser"
        self.parser = argparse.ArgumentParser(prog='PROG')
        self.parser.add_argument("--fin", action="store",
                                 dest="fin", default="", help="Input file with update fields (optional)")
        self.parser.add_argument("--dry-run", action="store_true",
                                 dest="dryrun", help="Fetch docs but do not update")
        self.parser.add_argument("--url", action="store",
                                 dest="url", default="https://cmsweb-testbed.cern.ch",
                                 help="URL for the MSPileup service")
        self.parser.add_argument("--userdn", action="store",
                                 dest="userDN", default='', help="user DN")
        self.parser.add_argument("--add-tran-record", action="store_true",
                                 dest="transition", help="add transition record")


class HTTPSClientAuthHandler(urllib.request.HTTPSHandler):
    """
    Basic HTTPS class
    """

    def __init__(self, key, cert):
        urllib.request.HTTPSHandler.__init__(self)
        # Create a default ssl context manager to carry the credentials.
        # It also loads the default CA certificates
        self.sslContext = ssl.create_default_context()
        self.sslContext.load_cert_chain(cert, keyfile=key)
        # Also load the default CA certificates (apparently not needed)
        # self.sslContext.load_verify_locations(None, '/etc/grid-security/certificates/')

    def https_open(self, req):
        """
        Return https opener
        :param req: request string
        """
        # Rather than pass in a reference to a connection class, we pass in
        # a reference to a function which, for all intents and purposes,
        # will behave as a constructor
        return self.do_open(self.getConnection, req)

    def getConnection(self, host, timeout=290):
        """
        Return HTTPSConnection object for given host
        :param host: host name
        :param timeout: timeout used in urllib.request call which will call this function
        """
        return HTTPSConnection(host, context=self.sslContext)


def userAgent():
    """
    Construct proper user agent string
    :return: user-agent string
    """
    major = sys.version_info.major
    minor = sys.version_info.minor
    clientId = f'updatePileupObjects::python/{major}.{minor}'
    user = os.environ.get('USER', '')
    client = f'{clientId} ({user})'
    return client


def getPileupDocs(mspileupUrl, handler, logger):
    """
    Fetch pileup documents from MSPileup
    :param mspileupUrl: string with the MSPileup url
    :param handler: HTTP handler for url opener
    :param logger: logger object
    :return: a list with the pileup documents
    """
    url = mspileupUrl + "/ms-pileup/data/pileup"
    logger.info("Fetching pileup documents from URL: %s", url)

    # place request
    req = urllib.request.Request(url)
    req.add_header('Accept', 'application/json')
    req.add_header('User-agent', userAgent())
    opener = urllib.request.build_opener(handler)
    urllib.request.install_opener(opener)
    with urllib.request.urlopen(req) as resp:
        logger.debug("Response: %s status %s", resp, resp.status)
        # check status ot HTTP response
        if resp.status != 200:
            msg = f"Bad response from server, status {resp.status}, reason {resp.reason}"
            raise RuntimeError(msg)

        # decode and load JSON data
        sdata = resp.read().decode('utf8')
        data = json.loads(sdata)

        logger.debug("Response: %s", data)
        if data and not data.get("result", []):
            msg = f"Failed to retrieve pileup documents. Error: {data}"
            raise RuntimeError(msg)
        return data["result"]


def writePileupDocs(mspileupUrl, puDocs, handler, logger):
    """
    Update each pileup document in MSPileup
    :param mspileupUrl: string with the MSPileup url
    :param puDocs: list with pileup documents
    :param handler: HTTP handler for url opener
    :param logger: logger object
    :return: None
    """
    url = mspileupUrl + "/ms-pileup/data/pileup"
    logger.info("Starting to update pileup documents against backend: %s", url)
    for doc in puDocs:
        logger.info("Updating pileup document: %s", doc.get("pileupName"))
        jdoc = json.dumps(doc)
        data = bytes(jdoc, 'utf-8')
        req = urllib.request.Request(url, data=data, method="PUT")
        req.add_header('Accept', 'application/json')
        req.add_header('Content-Type', 'application/json')
        req.add_header('User-agent', userAgent())

        opener = urllib.request.build_opener(handler)
        urllib.request.install_opener(opener)
        with urllib.request.urlopen(req) as resp:
            sdata = resp.read().decode('utf8')
            data = json.loads(sdata)
            logger.debug("Response: %s", data)
            if data and data.get("result", []):
                msg = f"Failed to inject pileup document for {doc['pileupName']}. Error: {data}"
                logger.critical(msg)


def addTransitionRecords(handler, url, userDN, logger, dryrun):
    """
    Helper function to add and update in place transition record for existing MSPileup records
    :param handler: HTTP handler to use
    :param url: url to use (string)
    :param userDN: string representing user DN
    :param logger: logger to use
    :param dryrun: option to run dry-run mode (boolean)
    :return: nothing
    """
    if not userDN:
        raise Exception("No userDN provided")
    puDocs = getPileupDocs(url, handler, logger)
    logger.info("Found %d documents in MSPileup", len(puDocs))
    recordsToUpdate = []
    for rec in puDocs:
        if not rec.get("transition"):
            tranRecord = {'containerFraction': 1.0,
                          'customDID': rec['pileupName'],
                          'updateTime': gmtimeSeconds(),
                          'DN': userDN}
            rec['transition'] = [tranRecord]
            logger.info("New pileup document is: %s", rec)
            recordsToUpdate.append(rec)

    if dryrun:
        logger.info("%s documents needs to be updated but are not written due to --dry-run option\n", len(recordsToUpdate))
    else:
        # finally, update the pileup documents in the database
        writePileupDocs(url, recordsToUpdate, handler, logger)
        logger.info("%s documents are uploaded to MSPileup\n", len(recordsToUpdate))


def updatePileupRecords(handler, url, logger, dryrun, fin=""):
    """
    helper function to update pileup records
    :param handler: HTTP handler to use
    :param url: url to use (string)
    :param logger: logger object to use
    :param dryrun: option to run dry-run mode (boolean)
    :param fin: input file name to read from (optional, string)
    :return: nothing
    """
    if fin:
        if os.path.exists(fin):
            with open(fin, 'r', encoding='utf-8') as istream:
                puOverride = json.load(istream)
            logger.info("Pileup override file: %s has the following content: %s", fin, puOverride)
        else:
            logger.error("Provided %s file does not exist", fin)
            os.exit(1)
    else:
        logger.info("No input file is provided to update pileup records")
        os.exit(1)

    puDocs = getPileupDocs(url, handler, logger)
    logger.info("Found %d documents in MSPileup", len(puDocs))

    # update the documents with the override content
    for doc in puDocs:
        doc.update(puOverride)
        logger.info("New pileup document is: %s", doc)

    if dryrun:
        logger.info("Updated documents (dry-run option):")
        for doc in puDocs:
            logger.info(doc)
    else:
        # finally, update the pileup documents in the database
        writePileupDocs(url, puDocs, handler, logger)
        logger.info("documents are uploaded to MSPileup\n")


def main():
    """Executes everything"""
    optmgr = OptionParser()
    opts = optmgr.parser.parse_args()
    logger = logging.getLogger('updatePileupObjects')
    logger.setLevel(logging.INFO)
    logging.basicConfig()

    # setup proxy/cert
    proxy = os.getenv('X509_USER_PROXY', '')
    if proxy:
        cert = proxy
        key = proxy
    else:
        cert = os.getenv('X509_USER_CERT', '')
        key = os.getenv('X509_USER_KEY', '')
        if not cert or not key:
            defaultCert = os.path.join(os.getenv('HOME'), '.globus/usercert.pem')
            defaultKey = os.path.join(os.getenv('HOME'), '.globus/userkey.pem')
            if os.path.isfile(defaultCert) and os.path.isfile(defaultKey):
                cert = defaultCert
                key = defaultKey

    # final check, the cert should be already set either by proxy or itself
    if not cert:
        logger.error("You need to define the X509_USER_CERT/X509_USER_key or X509_USER_PROXY env variables")
        sys.exit(1)
    logger.info("Using cert=%s key=%s", cert, key)

    # initialize HTTPS client handler
    handler = HTTPSClientAuthHandler(key, cert)

    if opts.transition:
        addTransitionRecords(handler, opts.url, opts.userDN, logger, opts.dryrun)
    else:
        updatePileupRecords(handler, opts.url, logger, opts.dryrun, opts.fin)


if __name__ == '__main__':
    main()



#!/usr/bin/env python
"""
_GetUnsubscribedBlocks_

MySQL implementation of RucioInjector.Database.GetUnsubscribedBlocks
"""
from __future__ import division

from WMCore.Database.DBFormatter import DBFormatter


class GetUnsubscribedBlocks(DBFormatter):
    """
    _GetUnsubscribedBlocks_

    Gets the unsubscribed blocks from DBSBuffer, so blocks
    without a valid rule id
    """
    sql = """SELECT dbsbuffer_block.blockname, dbsbuffer_location.pnn
               FROM dbsbuffer_block
             INNER JOIN dbsbuffer_location ON
                   dbsbuffer_block.location = dbsbuffer_location.id
             WHERE dbsbuffer_block.rule_id = '0'"""

    def execute(self, conn=None, transaction=False):
        result = self.dbi.processData(self.sql, conn=conn,
                                      transaction=transaction)
        return self.formatDict(result)



"""

AlertManagerAPI - send alerts to MONIT AlertManager via API calls
"""

from __future__ import division
from builtins import object
from datetime import timedelta, datetime
import socket
import json
import logging

from WMCore.Services.pycurl_manager import RequestHandler
from Utils.Timers import LocalTimezone
from Utils.Utilities import normalize_spaces
from WMCore.Services.UUIDLib import makeUUID


class AlertManagerAPI(object):
    """
    A class used to send alerts via the MONIT AlertManager API
    """

    def __init__(self, alertManagerUrl, logger=None):
        self.alertManagerUrl = alertManagerUrl
        # sender's hostname is added as an annotation
        self.hostname = socket.gethostname()
        self.mgr = RequestHandler()
        self.ltz = LocalTimezone()
        self.headers = {"Content-Type": "application/json"}
        self.validSeverity = ["high", "medium", "low"]
        self.logger = logger if logger else logging.getLogger()

    def sendAlert(self, alertName, severity, summary, description, service, tag="wmcore", endSecs=600, generatorURL=""):
        """
        :param alertName: a unique name for the alert
        :param severity: low, medium, high
        :param summary: a short description of the alert
        :param description: a longer informational message with details about the alert
        :param service: the name of the service firing an alert
        :param tag: a unique tag used to help route the alert
        :param endSecs: how many minutes until the alarm is silenced
        :param generatorURL: this URL will be sent to AlertManager and configured as a clickable "Source" link in the web interface

        AlertManager JSON format reference: https://www.prometheus.io/docs/alerting/latest/clients/
        [
          {
            "labels": {
            "alertname": "<requiredAlertName>",
            "<labelname>": "<labelvalue>",
            ...
          },
            "annotations": {
              "<labelname>": "<labelvalue>",
            ...
          },
            "startsAt": "<rfc3339>", # optional, will be current time if not present
            "endsAt": "<rfc3339>",
            "generatorURL": "<generator_url>" # optional
          },
        ]
        """

        if not self._isValidSeverity(severity):
            return False

        request = []
        alert = {}
        labels = {}
        annotations = {}

        # add labels
        labels["alertname"] = alertName
        labels["severity"] = severity
        labels["tag"] = tag
        labels["service"] = service
        labels["uuid"] = makeUUID()
        alert["labels"] = labels

        # add annotations
        annotations["hostname"] = self.hostname
        annotations["summary"] = normalize_spaces(summary)
        annotations["description"] = normalize_spaces(description)
        alert["annotations"] = normalize_spaces(annotations)

        # In python3 we won't need the LocalTimezone class
        # Will change to d = datetime.now().astimezone() + timedelta(seconds=endSecs)
        d = datetime.now(self.ltz) + timedelta(seconds=endSecs)
        alert["endsAt"] = d.isoformat("T")
        alert["generatorURL"] = generatorURL

        request.append(alert)
        # need to do this because pycurl_manager only accepts dict and encoded strings type
        params = json.dumps(request)

        # provide dump of alert send to AM which will allow to match it in WM logs
        header, res = self.mgr.request(self.alertManagerUrl, params=params, headers=self.headers, verb='POST')
        self.logger.info("ALERT: name=%s UUID=%s, HTTP status code=%s", alertName, labels["uuid"], header.status)

        return res

    def _isValidSeverity(self, severity):
        """
        Used to check if the severity of the alert matches the valid levels: low, medium, high
        :param severity: severity of the alert
        :return: True or False
        """
        if severity not in self.validSeverity:
            logging.critical("Alert submitted to AlertManagerAPI with invalid severity: %s", severity)
            return False
        return True



"""
_Repack_

Repack workflow

repacking -> RAW -> optional merge
             (supports multiple output with different primary datasets)

"""

from __future__ import division

from Utils.Utilities import makeList
from WMCore.Lexicon import procstringT0
from WMCore.WMSpec.StdSpecs.StdBase import StdBase


class RepackWorkloadFactory(StdBase):
    """
    _RepackWorkloadFactory_

    Stamp out Repack workflows.
    """

    def __init__(self):
        StdBase.__init__(self)

        self.inputPrimaryDataset = None
        self.inputProcessedDataset = None

        return

    def buildWorkload(self):
        """
        _buildWorkload_

        Build the workload given all of the input parameters.  At the very least
        this will create a processing task and merge tasks for all the outputs
        of the processing task.

        Not that there will be LogCollect tasks created for each processing
        task and Cleanup tasks created for each merge task.

        """
        workload = self.createWorkload()
        workload.setDashboardActivity("t0")

        cmsswStepType = "CMSSW"
        taskType = "Processing"

        # complete output configuration
        for output in self.outputs:
            moduleLabel = "write_%s_%s" % (output['primaryDataset'],
                                           output['dataTier'])
            output['moduleLabel'] = moduleLabel.replace("-", "_")  # For T0 Raw Skims, PDs will contain a "-", so here we replace for "_" for the moduleLabel

        # finalize splitting parameters
        mySplitArgs = self.repackSplitArgs.copy()
        mySplitArgs['algo_package'] = "T0.JobSplitting"

        repackTask = workload.newTask("Repack")
        repackOutMods = self.setupProcessingTask(repackTask, taskType,
                                                 scenarioName=self.procScenario,
                                                 scenarioFunc="repack",
                                                 scenarioArgs={'outputs': self.outputs,
                                                               'globalTag': self.globalTag},
                                                 splitAlgo="Repack",
                                                 splitArgs=mySplitArgs,
                                                 stepType=cmsswStepType)

        repackTask.setTaskType("Repack")

        self.addLogCollectTask(repackTask)

        for repackOutLabel in repackOutMods:
            self.addRepackMergeTask(repackTask, repackOutLabel)

        # setting the parameters which need to be set for all the tasks
        # sets acquisitionEra, processingVersion, processingString
        workload.setTaskPropertiesFromWorkload()

        # set the LFN bases (normally done by request manager)
        # also pass run number to add run based directories
        workload.setLFNBase(self.mergedLFNBase, self.unmergedLFNBase,
                            runNumber=self.runNumber)

        return workload

    def addRepackMergeTask(self, parentTask, parentOutputModuleName):
        """
        _addRepackMergeTask_

        Create an repackmerge task for files produced by the parent task.

        """
        mergeTask = parentTask.addTask("%sMerge%s" % (parentTask.name(), parentOutputModuleName))
        self.addRuntimeMonitors(mergeTask)
        mergeTaskCmssw = mergeTask.makeStep("cmsRun1")
        mergeTaskCmssw.setStepType("CMSSW")

        mergeTaskStageOut = mergeTaskCmssw.addStep("stageOut1")
        mergeTaskStageOut.setStepType("StageOut")
        mergeTaskLogArch = mergeTaskCmssw.addStep("logArch1")
        mergeTaskLogArch.setStepType("LogArchive")

        mergeTask.setTaskLogBaseLFN(self.unmergedLFNBase)

        self.addLogCollectTask(mergeTask, taskName="%s%sMergeLogCollect" % (parentTask.name(), parentOutputModuleName))

        mergeTask.applyTemplates()

        parentTaskCmssw = parentTask.getStep("cmsRun1")
        parentOutputModule = parentTaskCmssw.getOutputModule(parentOutputModuleName)
        dataTier = getattr(parentOutputModule, "dataTier")

        mergeTask.setInputReference(parentTaskCmssw, outputModule=parentOutputModuleName, dataTier=dataTier)

        mergeTaskCmsswHelper = mergeTaskCmssw.getTypeHelper()

        mergeTaskCmsswHelper.cmsswSetup(self.frameworkVersion, softwareEnvironment="",
                                        scramArch=self.scramArch)

        mergeTaskCmsswHelper.setErrorDestinationStep(stepName=mergeTaskLogArch.name())
        mergeTaskCmsswHelper.setGlobalTag(self.globalTag)
        mergeTaskCmsswHelper.setOverrideCatalog(self.overrideCatalog)

        # mergeTaskStageHelper = mergeTaskStageOut.getTypeHelper()
        # mergeTaskStageHelper.setMinMergeSize(0, 0)

        mergeTask.setTaskType("Merge")

        # finalize splitting parameters
        mySplitArgs = self.repackMergeSplitArgs.copy()
        mySplitArgs['algo_package'] = "T0.JobSplitting"

        mergeTask.setSplittingAlgorithm("RepackMerge",
                                        **mySplitArgs)
        mergeTaskCmsswHelper.setDataProcessingConfig(self.procScenario, "merge")

        self.addOutputModule(mergeTask, "Merged",
                             primaryDataset=getattr(parentOutputModule, "primaryDataset"),
                             dataTier=getattr(parentOutputModule, "dataTier"),
                             filterName=getattr(parentOutputModule, "filterName"),
                             rawSkim=getattr(parentOutputModule, "rawSkim", None),
                             forceMerged=True)

        self.addOutputModule(mergeTask, "MergedError",
                             primaryDataset=getattr(parentOutputModule, "primaryDataset") + "-Error",
                             dataTier=getattr(parentOutputModule, "dataTier"),
                             filterName=getattr(parentOutputModule, "filterName"),
                             rawSkim=getattr(parentOutputModule, "rawSkim", None),
                             forceMerged=True)

        self.addCleanupTask(parentTask, parentOutputModuleName, dataTier=getattr(parentOutputModule, "dataTier"))

        return mergeTask

    def __call__(self, workloadName, arguments):
        """
        _call_

        Create a Repack workload with the given parameters.
        """
        StdBase.__call__(self, workloadName, arguments)

        # Required parameters that must be specified by the Requestor.
        self.outputs = arguments['Outputs']

        # job splitting parameters
        self.repackSplitArgs = {}
        self.repackSplitArgs['maxSizeSingleLumi'] = arguments['MaxSizeSingleLumi']
        self.repackSplitArgs['maxSizeMultiLumi'] = arguments['MaxSizeMultiLumi']
        self.repackSplitArgs['maxInputEvents'] = arguments['MaxInputEvents']
        self.repackSplitArgs['maxInputFiles'] = arguments['MaxInputFiles']
        self.repackSplitArgs['maxLatency'] = arguments['MaxLatency']
        self.repackMergeSplitArgs = {}
        self.repackMergeSplitArgs['minInputSize'] = arguments['MinInputSize']
        self.repackMergeSplitArgs['maxInputSize'] = arguments['MaxInputSize']
        self.repackMergeSplitArgs['maxEdmSize'] = arguments['MaxEdmSize']
        self.repackMergeSplitArgs['maxOverSize'] = arguments['MaxOverSize']
        self.repackMergeSplitArgs['maxInputEvents'] = arguments['MaxInputEvents']
        self.repackMergeSplitArgs['maxInputFiles'] = arguments['MaxInputFiles']
        self.repackMergeSplitArgs['maxLatency'] = arguments['MaxLatency']

        return self.buildWorkload()

    @staticmethod
    def getWorkloadCreateArgs():
        baseArgs = StdBase.getWorkloadCreateArgs()
        specArgs = {"RequestType": {"default": "Repack"},
                    "ConfigCacheID": {"optional": True, "null": True},
                    "Scenario": {"default": "fake", "attr": "procScenario"},
                    "GlobalTag": {"default": "fake", "attr": "globalTag"},
                    "ProcessingString": {"default": "", "validate": procstringT0},
                    "Outputs": {"type": makeList, "optional": False},
                    "MaxSizeSingleLumi": {"type": int, "optional": False},
                    "MaxSizeMultiLumi": {"type": int, "optional": False},
                    "MaxInputEvents": {"type": int, "optional": False},
                    "MaxInputFiles": {"type": int, "optional": False},
                    "MaxLatency": {"type": int, "optional": False},
                    "MinInputSize": {"type": int, "optional": False},
                    "MaxInputSize": {"type": int, "optional": False},
                    "MaxEdmSize": {"type": int, "optional": False},
                    "MaxOverSize": {"type": int, "optional": False},
                    }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs

    @staticmethod
    def getWorkloadAssignArgs():
        baseArgs = StdBase.getWorkloadAssignArgs()
        specArgs = {
            "Override": {"default": {"eos-lfn-prefix": "root://eoscms.cern.ch//eos/cms/store/logs/prod/recent/Repack"},
                         "type": dict},
            }
        baseArgs.update(specArgs)
        StdBase.setDefaultArgumentsProperty(baseArgs)
        return baseArgs



#!/usr/bin/env python
"""
_HeritageLFNParent_

Oracle implementation of DBSBufferFiles.HeritageLFNParent
"""




from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.HeritageLFNParent import \
     HeritageLFNParent as MySQLHeritageLFNParent

class HeritageLFNParent(MySQLHeritageLFNParent):

    sql = """INSERT INTO dbsbuffer_file_parent (child, parent)
               VALUES (
                 (SELECT dfc.id FROM dbsbuffer_file dfc WHERE dfc.lfn = :child),
                 (SELECT dfp.id FROM dbsbuffer_file dfp WHERE dfp.lfn = :lfn)
               )"""



#!/usr/bin/env python
"""
DeleteMgr

Util class to provide delete functionality as an interface object.

Based on StageOutMgr class

"""
import logging

from builtins import object
from WMCore.Storage.Registry import retrieveStageOutImpl
from WMCore.Storage.RucioFileCatalog import storageJsonPath, readRFC
from WMCore.Storage.SiteLocalConfig import stageOutStr, loadSiteLocalConfig
from WMCore.Storage.StageOutError import StageOutFailure, StageOutInitError
from WMCore.WMException import WMException


class DeleteMgrError(WMException):
    """
    _DeleteMgrError_

    Specific exception class to work out file deletion exception details
    """

    def __init__(self, message, **data):
        WMException.__init__(self, message, **data)
        self.data.setdefault("ErrorCode", 60313)
        self.data.setdefault("ErrorType", self.__class__.__name__)


class DeleteMgr(object):
    """
    _DeleteMgr_

    Object that can be used to delete a set of files
    using TFC or an override.

    """

    def __init__(self, **overrideParams):

        self.logger = overrideParams.pop("logger", logging.getLogger())
        self.overrideConf = overrideParams

        # pairs of stageOut and Rucio file catalog: [(stageOut1,rfc1),(stageOut2,rfc2), ...]
        # a "stageOut" corresponds to a entry in the <stage-out> block in the site-local-config.xml, for example <method volume="KIT_dCache" protocol="WebDAV"/>
        # a "rfc" is the correponding RucioFileCatalog instance (RucioFileCatalog.py) of this "stageOut"
        self.stageOuts_rfcs = []
        self.numberOfRetries = 3
        self.retryPauseTime = 600

        #  //
        # // If override isnt None, we dont need SiteCfg, if it is
        # //  then we need siteCfg otherwise we are dead.

        if not self.overrideConf:
            self.siteCfg = loadSiteLocalConfig()
            self.initialiseSiteConf()
        else:
            self.initialiseOverride()

    def initialiseSiteConf(self):
        """
        _initialiseSiteConf_

        Extract required information from site conf and TFC

        """

        self.stageOuts = self.siteCfg.stageOuts

        self.logger.info("There are %s stage out definitions.\n" % len(self.stageOuts))

        for stageOut in self.stageOuts:
            foundNoneAttr = False
            for k in ['phedex-node', 'command', 'storageSite', 'volume', 'protocol']:
                v = stageOut.get(k)
                if v is None:
                    msg = "Unable to retrieve " + k + " of this stageOut: \n"
                    msg += stageOutStr(stageOut) + "\n"
                    msg += "from site config file.\n"
                    msg += "Continue to the next stageOut.\n"
                    self.logger.error(msg)
                    foundNoneAttr = True
                    break
            if foundNoneAttr: continue
            storageSite = stageOut.get("storageSite")
            volume = stageOut.get("volume")
            protocol = stageOut.get("protocol")
            command = stageOut.get("command")
            pnn = stageOut.get("phedex-node")

            self.logger.info("\tStage out to : %s using: %s \n" % (pnn, command))

            try:
                aPath = storageJsonPath(self.siteCfg.siteName, self.siteCfg.subSiteName, storageSite)
                rfc = readRFC(aPath, storageSite, volume, protocol)
                self.stageOuts_rfcs.append((stageOut, rfc))
                msg = "Rucio File Catalog has been loaded:\n"
                msg += str(self.stageOuts_rfcs[-1][1])
                self.logger.info(msg)
            except Exception as ex:
                msg = "Unable to load Rucio File Catalog:\n"
                msg += "This stage out will not be attempted:\n"
                msg += stageOutStr(stageOut) + '\n'
                msg += str(ex)
                self.logger.exception(msg)
                continue

        # no Rucio file catalog is initialized
        if not self.stageOuts_rfcs:
            raise StageOutInitError("===>Can not initialize Rucio file catalog")

        return

    def initialiseOverride(self):
        """
        _initialiseOverride_

        Extract and verify that the Override parameters are all present

        """

        overrideConf = {
            "command": None,
            "option": None,
            "phedex-node": None,
            "lfn-prefix": None,
            }

        try:
            overrideConf['command'] = self.overrideConf['command']
            overrideConf['phedex-node'] = self.overrideConf['phedex-node']
            overrideConf['lfn-prefix'] = self.overrideConf['lfn-prefix']
        except Exception as ex:
            msg = "Unable to extract override parameters from config:\n"
            msg += str(self.overrideConf)
            raise StageOutInitError(msg)
        if self.overrideConf.get('option') is not None:
            if len(self.overrideConf['option']) > 0:
                overrideConf['option'] = self.overrideConf['option']
            else:
                overrideConf['option'] = ""

        self.overrideConf = overrideConf

        msg = "=======Delete Override Initialised:================\n"
        for key, val in overrideConf.items():
            msg += " %s : %s\n" % (key, val)
        msg += "=====================================================\n"
        self.logger.info(msg)

        return

    def __call__(self, fileToDelete):
        """
        _operator()_

        Use call to delete a file

        """
        self.logger.info("==>Working on file: %s" % fileToDelete['LFN'])

        lfn = fileToDelete['LFN']

        deleteSuccess = False

        if not self.overrideConf:
            self.logger.info("===> Attempting to delete with %s stage outs", len(self.stageOuts))
            for stageOut_rfc in self.stageOuts_rfcs:
                if not deleteSuccess:
                    try:
                        fileToDelete['PNN'] = stageOut_rfc[0]['phedex-node']
                        fileToDelete['PFN'] = self.deleteLFN(lfn, stageOut_rfc)
                        deleteSuccess = True
                        break
                    except Exception as ex:
                        continue
        else:
            self.logger.info("===> Attempting stage outs from override")
            try:
                fileToDelete['PNN'] = self.overrideConf['phedex-node']
                fileToDelete['PFN'] = self.deleteLFN(lfn)
                deleteSuccess = True
            except Exception as ex:
                self.logger.error("===> Local file deletion failure. Exception:\n%s", str(ex))
        if deleteSuccess:
            msg = "===> Delete Successful:\n"
            msg += "====> LFN: %s\n" % fileToDelete['LFN']
            msg += "====> PFN: %s\n" % fileToDelete['PFN']
            msg += "====> PNN:  %s\n" % fileToDelete['PNN']
            self.logger.info(msg)
            return fileToDelete
        else:
            msg = "Unable to delete file:\n"
            msg += fileToDelete['LFN']
            raise StageOutFailure(msg, **fileToDelete)

    def deleteLFN(self, lfn, stageOut_rfc=None):
        """
        deleteLFN
        Given the lfn and an stageOut Rucio file catalog pair or override config, invoke the delete
        lfn: logical file name
        stageOut_rfc: a pair fo stageOut and correponding Rucio file catalog, required when no override provided 
        self.overrideConf: the follwoing params should be defined for override
            command - the stage out impl plugin name to be used
            option - the option values to be passed to that command (None is allowed)
            lfn-prefix - the LFN prefix to generate the PFN
            phedex-node - the Name of the PNN to which the file is being xferred
        """
        if not self.overrideConf:
            if stageOut_rfc is None:
                msg = "Can not delete lfn because of missing stage out information (stageOut_rfc is None): \n %s" % lfn
                raise StageOutFailure(msg, LFN=lfn)
            # FIXME there is circular import that is why this module is imported here
            from WMCore.Storage.StageOutMgr import searchRFC
            command = stageOut_rfc[0]['command']
            pfn = searchRFC(stageOut_rfc[1], lfn)
            if pfn is None:
                msg = "Unable to match lfn to pfn: \n  %s" % lfn
                raise StageOutFailure(msg, LFN=lfn, STAGEOUT=stageOutStr(stageOut_rfc[0]))
        else:
            if self.overrideConf['lfn-prefix'] is None:
                msg = "Unable to match lfn to pfn during lfn deletion because override lfn-prefix is None: \n %s" % lfn
                raise StageOutFailure(msg, LFN=lfn)
            command = self.overrideConf['command']
            pfn = "%s%s" % (self.overrideConf['lfn-prefix'], lfn)
        return self.deletePFN(pfn, lfn, command)

    def deletePFN(self, pfn, lfn, command):
        """
        Delete the given PFN
        """
        try:
            impl = retrieveStageOutImpl(command)
        except Exception as ex:
            msg = "Unable to retrieve impl for file deletion in %s\n" % (pfn)
            msg += "Error retrieving StageOutImpl for command named: %s\n" % (
                command,)
            raise StageOutFailure(msg, Command=command,
                                  LFN=lfn, ExceptionDetail=str(ex))
        impl.numRetries = self.numberOfRetries
        impl.retryPause = self.retryPauseTime

        try:
            impl.removeFile(pfn)
        except Exception as ex:
            self.logger.exception("Failed to delete file: %s", pfn)
            raise ex from None

        return pfn



"""
Oracle implementation of Add WorkUnit
"""
from __future__ import absolute_import, division, print_function

import time

from WMCore.WMBS.MySQL.WorkUnit.Add import Add as AddWorkUnitMySQL, MAX_EVENT


class Add(AddWorkUnitMySQL):
    """
    _Add_

    Use Oracle sequence instead of auto_increment and do it in a combined transaction
    """

    sql = """INSERT INTO wmbs_workunit (taskid, retry_count, last_unit_count, last_submit_time, status)
             VALUES (:taskid, :retry_count, :last_unit_count, :last_submit_time, :status)"""

    sql2 = """INSERT INTO wmbs_workunit (firstevent, lastevent, fileid, run, lumi)
             VALUES (:firstevent, :lastevent, :fileid, :run, :lumi)"""

    def execute(self, taskid=None, retry_count=0, last_unit_count=0, last_submit_time=time.time(),
                status=0, first_event=0, last_event=MAX_EVENT,
                fileid=None, run=None, lumi=None,
                conn=None, transaction=False):
        """
        Add the parts of the object that go into workunit and into the association table
        """

        if taskid:
            binds = {'taskid': taskid, 'retry_count': retry_count, 'last_unit_count': last_unit_count,
                     'last_submit_time': last_submit_time, 'status': status,
                     'firstevent': first_event, 'lastevent': last_event, 'fileid': fileid, 'run': run, 'lumi': lumi}

            self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)
            self.dbi.processData(self.sql2, binds, conn=conn, transaction=transaction)
        else:
            raise RuntimeError("No way to create work units without task ID")

        return



"""
API for querying the status of agent drain process
"""

from __future__ import division
from builtins import object

from WMComponent.DBS3Buffer.DBSBufferUtil import DBSBufferUtil
from WMCore.Services.PyCondor.PyCondorAPI import PyCondorAPI
from WMCore.WorkQueue.WorkQueueBackend import WorkQueueBackend


class DrainStatusAPI(object):
    """
    Provides methods for querying dbs and condor for drain statistics
    """
    def __init__(self, config):
        # queue url used in WorkQueueManager
        self.thisAgentUrl = "http://" + config.Agent.hostName + ":5984"
        self.globalBackend = WorkQueueBackend(config.WorkloadSummary.couchurl)
        self.localBackend = WorkQueueBackend(config.WorkQueueManager.couchurl)
        self.dbsUtil = DBSBufferUtil()
        self.condorAPI = PyCondorAPI()
        self.condorStates = ("Running", "Idle")

    def collectDrainInfo(self):
        """
        Call methods to check the drain status
        """
        results = {}
        results['workflows_completed'] = self.checkWorkflows()

        # if workflows are completed, collect additional drain statistics
        if results['workflows_completed']:
            results['upload_status'] = self.checkFileUploadStatus()
            results['condor_status'] = self.checkCondorStates()
            results['local_wq_status'] = self.checkLocalWQStatus(dbname="workqueue")
            results['local_wqinbox_status'] = self.checkLocalWQStatus(dbname="workqueue_inbox")
            results['global_wq_status'] = self.checkGlobalWQStatus()

        return results

    def checkWorkflows(self):
        """
        Check to see if all workflows have a 'completed' status
        """
        results = self.dbsUtil.isAllWorkflowCompleted()
        return results

    def checkCondorStates(self):
        """
        Check idle and running jobs in Condor
        """
        results = {}
        jobs = self.condorAPI.getCondorJobsSummary()
        for state in self.condorStates:
            # if there is an error, report it instead of the length of an empty list
            if not jobs:
                results[state.lower()] = None
            else:
                results[state.lower()] = int(jobs[0].get(state))

        return results

    def checkFileUploadStatus(self):
        """
        Check file upload status:
            Blocks open in DBS
            Files not uploaded in DBS
            Files not uploaded to Phedex
        """
        results = {}
        results['dbs_open_blocks'] = self.dbsUtil.countOpenBlocks()
        results['dbs_notuploaded'] = self.dbsUtil.countFilesByStatus(status="NOTUPLOADED")
        results['phedex_notuploaded'] = self.dbsUtil.countPhedexNotUploaded()
        return results

    def checkLocalWQStatus(self, dbname):
        """
        Query local WorkQueue workqueue/workqueue_inbox database to see whether
        there are any active elements in this agent.
        """
        results = {}

        for st in ('Available', 'Negotiating', 'Acquired', 'Running'):
            if dbname == "workqueue":
                elements = self.localBackend.getElements(status=st, returnIdOnly=True)
            else:
                elements = self.localBackend.getInboxElements(status=st, returnIdOnly=True)
            results[st] = len(elements)
        return results

    def checkGlobalWQStatus(self):
        """
        Query Global WorkQueue workqueue database to see whether there are
        any active elements set to this agent.
        """
        results = {}

        for st in ("Acquired", "Running"):
            elements = self.globalBackend.getElements(status=st, returnIdOnly=True,
                                                      ChildQueueUrl=self.thisAgentUrl)
            results[st] = len(elements)
        return results



"""
_WorkUnit.GetByID_

Oracle implementation of WorkUnit.GetByID
"""

from __future__ import absolute_import, division, print_function

from WMCore.WMBS.MySQL.WorkUnit.GetByID import GetByID as GetByIDMySQL


class GetByID(GetByIDMySQL):
    """
    _WorkUnit.GetByID_

    Oracle implementation of WorkUnit.GetByID
    """

    pass



#!/usr/bin/env python
"""
_AddLocation_

Oracle implementation of DBSBufferFiles.AddLocation
"""

from WMComponent.DBS3Buffer.MySQL.DBSBufferFiles.AddLocation import AddLocation as MySQLAddLocation

class AddLocation(MySQLAddLocation):

    sql = """INSERT INTO dbsbuffer_location (pnn)
             SELECT :location FROM dual
             WHERE NOT EXISTS (SELECT id FROM dbsbuffer_location WHERE pnn = :location)"""



#!/usr/bin/env python
"""
_WorkUnit.Delete_

MySQL implementation of WorkUnit.Delete
"""

from __future__ import absolute_import, division, print_function

from WMCore.Database.DBFormatter import DBFormatter


class Delete(DBFormatter):
    """
    _WorkUnit.Delete_

    MySQL implementation of WorkUnit.Delete
    """

    sql = 'DELETE FROM wmbs_workunit WHERE id = :id'

    def getBinds(self, name=None):
        return self.dbi.buildbinds(self.dbi.makelist(name), 'id')

    def execute(self, wuid=None, conn=None, transaction=False):
        self.dbi.processData(self.sql, self.getBinds(wuid), conn=conn, transaction=transaction)
        return True



#!/usr/bin/env python
"""
_Delete_

Oracle implementation of Jobs.Delete

"""

__all__ = []



from WMCore.WMBS.MySQL.Jobs.Delete import Delete as DeleteJobMySQL

class Delete(DeleteJobMySQL):
    sql = DeleteJobMySQL.sql



#!/usr/bin/env python
# encoding: utf-8
"""
CouchUtils.py

Created by Dave Evans on 2010-03-11.
Copyright (c) 2010 Fermilab. All rights reserved.
"""
from __future__ import print_function

from future import standard_library
standard_library.install_aliases()

from http.client import HTTPException

import functools

import WMCore.Database.CMSCouch as CMSCouch


class CouchConnectionError(Exception):
    """docstring for CouchConnectionError"""
    def __init__(self, arg):
        super(CouchConnectionError, self).__init__(arg)




def initialiseCouch(objectRef):
    if objectRef.couchdb != None:
        return
    if objectRef.url == None:
        msg = "url for couch service not provided"
        raise CouchConnectionError(msg)
    if objectRef.database == None:
        msg = "database name for couch service not provided"
        raise CouchConnectionError(msg)
    try:
        objectRef.server = CMSCouch.CouchServer(objectRef.url)
        objectRef.couchdb = objectRef.server.connectDatabase(objectRef.database)
    except HTTPException as e:
        msg = "%s with status: %s, reason: %s and result: %s" % (repr(e),
                                                                 getattr(e, 'status', ""),
                                                                 getattr(e, 'reason', ""),
                                                                 getattr(e, 'result', ""))
        raise CouchConnectionError(msg)
    except Exception as e:
        msg = "Exception instantiating couch services for :\n"
        msg += " url = %s\n database = %s\n" % (objectRef.url, objectRef.database)
        msg += " Exception: %s" % str(e)
        print(msg)
        raise CouchConnectionError(msg)

def connectToCouch(funcRef):
    """
    _connectToCouch_

    Decorator method to connect the function's class object to couch
    """
    @functools.wraps(funcRef)
    def wrapper(x, *args, **opts):
        initialiseCouch(x)
        return funcRef(x, *args, **opts)
    return wrapper

def requireOwner(func):
    """
    _requireOwner_

    Decorator to ensure that the owner attribute of a couch ACDC object is not None
    """

    def wrapper(self, *args, **opts):
        if self.owner == None:
            msg = "Owner not provided for Collection"
            raise RuntimeError(msg)
        return func(self, *args, **opts)
    return wrapper

def requireCollection(func):
    """
    _requireCollection_

    Decorator to ensure that the collection attribute of a couch ACDC object is not None
    """

    def wrapper(self, *args, **opts):
        if self.collection == None:
            msg = "Collection not provided for Collection"
            raise RuntimeError(msg)
        return func(self, *args, **opts)
    return wrapper

def requireFilesetName(func):
    """
    _requireFilesetName_

    Decorator to require that a fileset has a name that is not None

    """
    def wrapper(self, *args, **opts):
        if not 'name' in self or self['name'] == None:
            raise RuntimeError("Filesets must be named")
        return func(self, *args, **opts)
    return wrapper

def requireCollectionName(func):
    """
    _requireCollectionName_

    Decorator to require that a collection has a name that is not None

    """
    def wrapper(self, *args, **opts):
        if not 'name' in self or self['name'] == None:
            raise RuntimeError("Filesets must be named")
        return func(self, *args, **opts)
    return wrapper



#!/usr/bin/env python
"""
_File_

A simple object representing a file in WMBS.
"""

from builtins import next, str, bytes

import logging
import threading

from WMCore.DataStructs.File import File as WMFile
from WMCore.DataStructs.Run import Run
from WMCore.WMBS.WMBSBase import WMBSBase


class File(WMBSBase, WMFile):
    """
    A simple object representing a file in WMBS
    """

    def __init__(self, lfn="", id=-1, size=0, events=0, checksums={},
                 parents=None, locations=None, first_event=0,
                 last_event=0, merged=True):
        WMBSBase.__init__(self)
        WMFile.__init__(self, lfn=lfn, size=size, events=events,
                        checksums=checksums, parents=parents, merged=merged)

        if locations is None:
            self.setdefault("newlocations", set())
        else:
            if isinstance(locations, (str, bytes)):
                self.setdefault("newlocations", set())
                self['newlocations'].add(locations)
            else:
                self.setdefault("newlocations", locations)

        # overwrite the default value set from the WMFile
        self["first_event"] = first_event
        self["last_event"] = last_event
        self.setdefault("id", id)
        self['locations'] = set()

    def exists(self):
        """
        if id is exist (not -1) check with id first
        Does a file exist with this lfn, return the id
        """
        if self["id"] != -1:
            action = self.daofactory(classname="Files.ExistsByID")
            result = action.execute(id=self["id"], conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        else:
            action = self.daofactory(classname="Files.Exists")
            result = action.execute(lfn=self["lfn"], conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
            if result:
                self["id"] = result

        return result

    def getInfo(self):
        """
        Return the files attributes as a tuple
        """
        return (self['lfn'], self['id'], self['size'], self['events'],
                self['checksums'], list(self['runs']), list(self['locations']),
                list(self['parents']))

    def getLocations(self):
        """
        Get a list of locations for this file
        """
        return list(self['locations'])

    def getRuns(self):
        """
        Get a list of run lumi objects (List of set() of type
        WMCore.DataStructs.Run)
        """
        return list(self['runs'])

    def getParentLFNs(self):
        """
        Get a flat list of parent LFNs
        """
        result = []
        parents = self['parents']
        while parents:
            result.extend(parents)
            temp = []
            for parent in parents:
                temp.extend(parent["parents"])
            parents = temp
        result.sort()  # ensure SecondaryInputFiles are in order
        return [x['lfn'] for x in result]

    def getAncestors(self, level=2, type="id"):
        """
        Get ancestorLFNs. it will access directly DAO.
        level indicates the level of ancestors. default value is 2
        (grand parents). level should be bigger than >= 1
        """
        existingTransaction = self.beginTransaction()

        def _getAncestorIDs(ids, level):
            action = self.daofactory(classname="Files.GetParentIDsByID")
            parentIDs = sorted(action.execute(ids, conn=self.getDBConn(), transaction=self.existingTransaction()))

            if level == 1 or len(parentIDs) == 0:
                return parentIDs
            else:
                return _getAncestorIDs(parentIDs, level - 1)

        if self["id"] < 0:
            self.load()

        idList = _getAncestorIDs(self["id"], level)

        if type == "id":
            results = idList
        elif type == "lfn":
            results = []
            for fileID in idList:
                anceFile = File(id=fileID)
                anceFile.load()
                results.append(anceFile["lfn"])
        elif type == "file":
            results = []
            for fileID in idList:
                anceFile = File(id=fileID)
                anceFile.load()
                results.append(anceFile)

        self.commitTransaction(existingTransaction)
        return results

    def getDescendants(self, level=2, type="id"):
        """
        Get descendants. it will access directly DAO.
        level indicates the level of ancestors. default value is 2
        (grand parents). level should be bigger than >= 1
        """
        existingTransaction = self.beginTransaction()

        def _getDescendantIDs(ids, level):
            action = self.daofactory(classname="Files.GetChildIDsByID")
            childIDs = sorted(action.execute(ids, conn=self.getDBConn(), transaction=self.existingTransaction()))

            if level == 1 or len(childIDs) == 0:
                return childIDs
            else:
                return _getDescendantIDs(childIDs, level - 1)

        if self["id"] < 0:
            self.load()

        idList = _getDescendantIDs(self["id"], level)

        if type == "id":
            results = idList
        elif type == "lfn":
            results = []
            for fileID in idList:
                descFile = File(id=fileID)
                descFile.load()
                results.append(descFile['lfn'])
        elif type == "file":
            results = []
            for fileID in idList:
                descFile = File(id=fileID)
                descFile.load()
                results.append(descFile)

        self.commitTransaction(existingTransaction)
        return results

    def load(self):
        """
        _load_

        Load any meta data that is associated with a file.  This currently
        includes id, lfn, size, events and cksum.
        """
        if self["id"] > 0:
            action = self.daofactory(classname="Files.GetByID")
            result = action.execute(self["id"], conn=self.getDBConn(),
                                    transaction=self.existingTransaction())
        else:
            action = self.daofactory(classname="Files.GetByLFN")
            result = action.execute(self["lfn"], conn=self.getDBConn(),
                                    transaction=self.existingTransaction())

        self.update(result)
        self.loadChecksum()
        return

    def loadData(self, parentage=0):
        """
        _loadData_

        Load all information about a file.  This currently includes meta data,
        the run and lumi information, all the locations where the file
        is stored and any parentage information.  The parentage parameter to
        this method will determine how many generations to load.
        """
        existingTransaction = self.beginTransaction()

        if self["id"] < 0 or self["lfn"] == "":
            self.load()

        action = self.daofactory(classname="Files.GetRunLumiFile")
        runs = action.execute(self["lfn"], conn=self.getDBConn(),
                              transaction=self.existingTransaction())

        [self.addRun(run=Run(r, *runs[r])) for r in runs]

        action = self.daofactory(classname="Files.GetLocation")
        self["locations"] = action.execute(self["lfn"], conn=self.getDBConn(),
                                           transaction=self.existingTransaction())
        self["newlocations"].clear()
        self["parents"] = set()

        if parentage > 0:
            action = self.daofactory(classname="Files.GetParents")
            parentLFNs = action.execute(self["lfn"], conn=self.getDBConn(),
                                        transaction=self.existingTransaction())

            for lfn in parentLFNs:
                f = File(lfn=lfn)
                f.load()
                f.loadData(parentage=parentage - 1)
                self["parents"].add(f)

        self.commitTransaction(existingTransaction)
        return

    def create(self):
        """
        _create_

        Create a file.  If no transaction is passed in this will wrap all
        statements in a single transaction.
        """
        existingTransaction = self.beginTransaction()

        if self.exists():
            self.commitTransaction(existingTransaction)
            self.load()
            # assume if the file already exist, parentage is already set.
            # or not exist yet
            return

        addAction = self.daofactory(classname="Files.Add")
        addAction.execute(files=self["lfn"], size=self["size"],
                          events=self["events"],
                          first_event=self["first_event"],
                          merged=self["merged"],
                          conn=self.getDBConn(),
                          transaction=self.existingTransaction())

        if len(self["runs"]) > 0:
            lumiAction = self.daofactory(classname="Files.AddRunLumi")
            lumiAction.execute(file=self["lfn"], runs=self["runs"],
                               conn=self.getDBConn(),
                               transaction=self.existingTransaction())
        self.load()
        self.updateLocations()
        # call it here to make sure self["id"] exist
        if self["parents"]:
            for parent in self['parents']:
                parent.create()
                action = self.daofactory(classname="Files.Heritage")
                action.execute(child=self["id"], parent=parent["id"],
                               conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        if self['checksums']:
            # Add a checksum
            for entry in self['checksums']:
                self.setCksum(cksum=self['checksums'][entry], cktype=entry)
        return

    def delete(self):
        """
        Remove a file from WMBS
        """
        action = self.daofactory(classname="Files.Delete")
        action.execute(file=self["lfn"], conn=self.getDBConn(),
                       transaction=self.existingTransaction())
        return

    def addChild(self, lfn):
        """
        set an existing file (lfn) as a child of this file
        """
        existingTransaction = self.beginTransaction()

        child = File(lfn=lfn)
        child.load()

        if not self['id'] > 0:
            raise Exception("Parent file doesn't have an id %s" % self['lfn'])
        if not child['id'] > 0:
            raise Exception("Child file doesn't have an id %s" % child['lfn'])

        heritageAction = self.daofactory(classname="Files.Heritage")
        heritageAction.execute(child=child["id"], parent=self["id"],
                               conn=self.getDBConn(),
                               transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def addParent(self, lfn):
        """
        set an existing file (lfn) as a parent of this file
        """
        existingTransaction = self.beginTransaction()

        parent = File(lfn=lfn)
        parent.load()
        self["parents"].add(parent)

        if not self["id"] > 0:
            raise Exception("Child file doesn't have an id %s" % self["lfn"])
        if not parent["id"] > 0:
            raise Exception("Parent file doesn't have an id %s" % parent["lfn"])

        action = self.daofactory(classname="Files.Heritage")
        action.execute(child=self["id"], parent=parent["id"],
                       conn=self.getDBConn(),
                       transaction=self.existingTransaction())

        self.commitTransaction(existingTransaction)
        return

    def addRunSet(self, runSet):
        """
        add the set of runs.  This should be called after a file is created,
        unlike addRun which should be called before the file was created.
        runset should be set of DataStruct.Run
        also there should be no duplicate entries in runset.
        (May need to change in schema level not to allow duplicate record)
        """
        existingTransaction = self.beginTransaction()

        lumiAction = self.daofactory(classname="Files.AddRunLumi")
        lumiAction.execute(file=self["lfn"], runs=runSet,
                           conn=self.getDBConn(),
                           transaction=self.existingTransaction())

        action = self.daofactory(classname="Files.GetRunLumiFile")
        runs = action.execute(self["lfn"], conn=self.getDBConn(),
                              transaction=self.existingTransaction())

        self["runs"].clear()
        [self.addRun(run=Run(r, *runs[r])) for r in runs]

        self.commitTransaction(existingTransaction)
        return

    def updateLocations(self):
        """
        _updateLocations_

        Write any new locations to the database.  After any new locations are
        written to the database all locations will be reloaded from the
        database.
        """
        if not self.exists():
            return

        existingTransaction = self.beginTransaction()

        # Add new locations if required
        if len(self["newlocations"]) > 0:
            addAction = self.daofactory(classname="Files.SetLocation")
            addAction.execute(file=self["id"], pnn=self["newlocations"],
                              conn=self.getDBConn(),
                              transaction=self.existingTransaction())

        # Update locations from the DB
        getAction = self.daofactory(classname="Files.GetLocation")
        self["locations"] = getAction.execute(self["lfn"], conn=self.getDBConn(),
                                              transaction=self.existingTransaction())
        self["newlocations"].clear()

        self.commitTransaction(existingTransaction)
        return

    def setLocation(self, pnn, immediateSave=True):
        """
        Sets the location of a file. If immediateSave is True, commit change to
        the DB immediately, otherwise queue for addition when save() is called.
        Also removes previous error where a file would have to be saved before
        locations could be added - confusing when file requires locations on its
        first creation (breaks transaction model in Fileset commits etc)
        """
        if isinstance(pnn, (str, bytes)):
            self['newlocations'].add(pnn)
            self['locations'].add(pnn)
        else:
            self['newlocations'].update(pnn)
            self['locations'].update(pnn)

        if immediateSave:
            self.updateLocations()

        return

    def setCksum(self, cksum, cktype):
        """
        _setCKType_

        Set the Checksum Type
        """

        myThread = threading.currentThread()

        if self['id'] < 0:
            # You haven't bothered to create the file!
            return

        existingTransaction = self.beginTransaction()

        action = self.daofactory(classname="Files.AddChecksum")
        action.execute(fileid=self['id'], cktype=cktype, cksum=cksum,
                       conn=self.getDBConn(), transaction=existingTransaction)

        self.commitTransaction(existingTransaction)

        return

    def loadChecksum(self):
        """
        _loadChecksum_

        Grab checksums.  If plural, put a list of dictionaries of type
        {'cktype', 'cksum'} in both self['cktype'] and self['cksum']
        """
        # Now get the checksum
        action = self.daofactory(classname='Files.GetChecksum')
        result = action.execute(fileid=self['id'], conn=self.getDBConn(),
                                transaction=self.existingTransaction())
        if result:
            self.update(result)

        return

    def loadFromDataStructsFile(self, file):
        """
        _loadFromDataStructsFile_

        This function will create a WMBS File given a DataStructs file
        """
        self.update(file)
        # Clear the parents since update
        # will update the parents with set of lfns,
        # parents should be set of wmbs files in WMBS File class
        self["parents"] = set()

        if isinstance(file["locations"], set):
            s = file["locations"].copy()
            pnn = s.pop()
        elif isinstance(file["locations"], list):
            pnn = file["locations"][0]
        else:
            pnn = file["locations"]

        self.setLocation(pnn=pnn, immediateSave=False)

        self.create()

        for parent in file['parents']:
            self.addParent(parent)

        return

    def returnDataStructsFile(self):
        """
        _returnDataStructsFile_

        Creates a dataStruct file out of this file
        """
        parents = set()
        for parent in self["parents"]:
            parents.add(WMFile(lfn=parent['lfn'], size=parent['size'],
                               events=parent['events'], checksums=parent['checksums'],
                               parents=parent['parents'], merged=parent['merged']))

        file = WMFile(lfn=self['lfn'], size=self['size'],
                      events=self['events'], checksums=self['checksums'],
                      parents=parents, merged=self['merged'])

        for run in self['runs']:
            file.addRun(run)

        for location in self['locations']:
            file.setLocation(pnn=location)

        return file


def addFilesToWMBSInBulk(filesetId, workflowName, files, isDBS=True,
                         conn=None, transaction=None):
    """
    _addFilesToWMBSInBulk

    Do a bulk addition of files into WMBS. This is a speedup.

    Assumes files are full dao objects
    """
    if not files:
        # Nothing to do
        return 0

    daofactory = next(iter(files)).daofactory
    setParentage = daofactory(classname="Files.SetParentage")
    setFileRunLumi = daofactory(classname="Files.AddRunLumi")
    setFileLocation = daofactory(classname="Files.SetLocationForWorkQueue")
    setFileAddChecksum = daofactory(classname="Files.AddChecksumByLFN")
    addFileAction = daofactory(classname="Files.Add")
    addToFileset = daofactory(classname="Files.AddDupsToFileset")
    updateFileAction = daofactory(classname="Files.Update")

    # build up list of binds for all files then run in single transaction
    parentageBinds = []
    runLumiBinds = []
    fileCksumBinds = []
    fileLocations = []
    fileCreate = []
    fileLFNs = set()
    lfnsToCreate = set()
    lfnList = set()
    fileUpdate = []

    for wmbsFile in files:
        lfn = wmbsFile['lfn']
        lfnList.add(lfn)

        if wmbsFile.get('inFileset', True):
            fileLFNs.add(lfn)
        for parent in wmbsFile['parents']:
            parentageBinds.append({'child': lfn, 'parent': parent["lfn"]})

        selfChecksums = wmbsFile['checksums']
        if len(wmbsFile['runs']) > 0:
            runLumiBinds.append({'lfn': lfn, 'runs': wmbsFile['runs']})

        if len(wmbsFile['newlocations']) < 1:
            # Then we're in trouble
            msg = "File created in WMBS without locations!\n"
            msg += "File lfn: %s\n" % (lfn)
            logging.error(msg)
            raise RuntimeError(msg)

        for loc in wmbsFile['newlocations']:
            fileLocations.append({'lfn': lfn, 'location': loc})

        if wmbsFile.exists():
            # update events, size, first_event, merged
            fileUpdate.append([lfn,
                               wmbsFile['size'],
                               wmbsFile['events'],
                               None,
                               wmbsFile["first_event"],
                               wmbsFile['merged']])
            continue

        lfnsToCreate.add(lfn)

        if selfChecksums:
            # If we have checksums we have to create a bind
            # For each different checksum
            for entry in selfChecksums:
                fileCksumBinds.append({'lfn': lfn, 'cksum': selfChecksums[entry],
                                       'cktype': entry})

        fileCreate.append([lfn,
                           wmbsFile['size'],
                           wmbsFile['events'],
                           None,
                           wmbsFile["first_event"],
                           wmbsFile['merged']])

    if len(fileCreate) > 0:
        addFileAction.execute(files=fileCreate,
                              conn=conn,
                              transaction=transaction)
    if fileCksumBinds:
        setFileAddChecksum.execute(bulkList=fileCksumBinds,
                                   conn=conn,
                                   transaction=transaction)

    if len(fileUpdate) > 0:
        updateFileAction.execute(files=fileUpdate,
                                 conn=conn,
                                 transaction=transaction)

    if len(fileLocations) > 0:
        setFileLocation.execute(lfns=lfnList, locations=fileLocations,
                                isDBS=isDBS,
                                conn=conn,
                                transaction=transaction)
    if len(runLumiBinds) > 0:
        setFileRunLumi.execute(file=runLumiBinds,
                               conn=conn,
                               transaction=transaction)

    if len(fileLFNs) > 0:
        addToFileset.execute(file=fileLFNs,
                             fileset=filesetId,
                             workflow=workflowName,
                             conn=conn,
                             transaction=transaction)

    if len(parentageBinds) > 0:
        setParentage.execute(binds=parentageBinds,
                             conn=conn,
                             transaction=transaction)

    return len(lfnsToCreate)



#!/usr/bin/env python
"""
_StageOutMgr_

Util class to provide stage out functionality as an interface object.

Based of RuntimeStageOut.StageOutManager, that should probably eventually
use this class as a basic API
"""
import logging

from builtins import object
from future.utils import viewitems
from WMCore.Storage.DeleteMgr import DeleteMgr
from WMCore.Storage.Registry import retrieveStageOutImpl
from WMCore.Storage.RucioFileCatalog import storageJsonPath, readRFC
from WMCore.Storage.SiteLocalConfig import stageOutStr
from WMCore.Storage.StageOutError import StageOutFailure
from WMCore.Storage.StageOutError import StageOutInitError
from WMCore.WMException import WMException
# If we don't import them, they cannot be ever used (bad PyCharm!)
import WMCore.Storage.Backends
import WMCore.Storage.Plugins


def stageoutPolicyReport(fileToStage, pnn, command, stageOutType, stageOutExit):
    """
    Prepare some extra information regarding the stage out step (for both prod/analysis jobs).

    NOTE: this information used to be shipped to the old SSB dashboard. I'm unsure
    whether it's provided to any other monitoring system at the moment.
    """
    tempDict = {}
    tempDict['LFN'] = fileToStage['LFN'] if 'LFN' in fileToStage else None
    tempDict['PNN'] = fileToStage['PNN'] if 'PNN' in fileToStage else None
    tempDict['PNN'] = pnn if pnn else tempDict['PNN']
    tempDict['StageOutCommand'] = fileToStage['command'] if 'command' in fileToStage else None
    tempDict['StageOutCommand'] = command if command else tempDict['StageOutCommand']
    tempDict['StageOutType'] = stageOutType
    tempDict['StageOutExit'] = stageOutExit
    fileToStage['StageOutReport'].append(tempDict)
    return fileToStage


def searchRFC(rfc, lfn):
    """
    _searchRFC_

    Search the Rucio File Catalog for the lfn provided,
    if a match is made, return the matched PFN

    """
    if rfc == None:
        msg = "Rucio File Catalog not available to match LFN:\n"
        msg += lfn
        logging.error(msg)
        return None
    if rfc.preferredProtocol == None:
        msg = "Rucio File Catalog: " + str(rfc) + "does not have a preferred protocol\n"
        msg += "which prevents stage out for:\n"
        msg += lfn
        logging.error(msg)
        return None

    pfn = rfc.matchLFN(rfc.preferredProtocol, lfn)
    if pfn == None:
        msg = "Unable to map LFN to PFN:\n"
        msg += "LFN: %s\n" % lfn
        msg += "using this Rucio File Catalog\n"
        msg += str(rfc)
        logging.error(msg)
        return None

    msg = "LFN to PFN match made:\n"
    msg += "LFN: %s\nPFN: %s\n" % (lfn, pfn)
    logging.info(msg)

    return pfn


class StageOutMgr(object):
    """
    _StageOutMgr_

    Object that can be used to stage out a set of files
    using RFC or an override.

    """

    def __init__(self, **overrideParams):
        logging.info("StageOutMgr::__init__()")
        self.overrideConf = overrideParams

        # Figure out if any of the override parameters apply to stage-out
        self.override = False
        if overrideParams != {}:
            logging.info("StageOutMgr::__init__(): Override: %s", overrideParams)
            checkParams = ["command", "option", "phedex-node", "lfn-prefix"]
            for param in checkParams:
                if param in self.overrideConf:
                    self.override = True
            if not self.override:
                logging.info("=======StageOut Override: These are not the parameters you are looking for")

        self.substituteGUID = True

        self.stageOuts_rfcs = []  # pairs of stageOut and Rucio file catalog

        self.numberOfRetries = 3
        self.retryPauseTime = 600

        from WMCore.Storage.SiteLocalConfig import loadSiteLocalConfig

        #  //
        # // If override isnt None, we dont need SiteCfg, if it is
        # //  then we need siteCfg otherwise we are dead.

        if self.override == False:
            self.siteCfg = loadSiteLocalConfig()
            self.initialiseSiteConf()
        else:
            self.initialiseOverride()

        self.failed = {}
        self.completedFiles = {}
        return

    def initialiseSiteConf(self):
        """
        _initialiseSiteConf_

        Extract required information from site conf and RFC

        """
        logging.info("==== Stageout configuration start ====")

        self.stageOuts = self.siteCfg.stageOuts

        logging.info("\nThere are %s stage out definitions." % len(self.stageOuts))
        for stageOut in self.stageOuts:
            foundNoneAttr = False
            for k in ['phedex-node', 'command', 'storageSite', 'volume', 'protocol']:
                v = stageOut.get(k)
                if v is None:
                    amsg = ""
                    amsg += "Unable to retrieve " + k + " of this stageOut: \n"
                    amsg += stageOutStr(stageOut) + "\n"
                    amsg += "From site config file.\n"
                    amsg += "Continue to the next stageOut.\n"
                    logging.error(amsg)
                    foundNoneAttr = True
                    break
            if foundNoneAttr: continue
            storageSite = stageOut.get("storageSite")
            volume = stageOut.get("volume")
            protocol = stageOut.get("protocol")
            command = stageOut.get("command")
            pnn = stageOut.get("phedex-node")

            logging.info("\nStage out to : %s using: %s" % (pnn, command))

            try:
                aPath = storageJsonPath(self.siteCfg.siteName, self.siteCfg.subSiteName, storageSite)
                rfc = readRFC(aPath, storageSite, volume, protocol)
                self.stageOuts_rfcs.append((stageOut, rfc))
                msg = "\nRucio File Catalog has been loaded:"
                msg += "\n" + str(rfc)
                logging.info(msg)
            except Exception as ex:
                amsg = "\nUnable to load Rucio File Catalog. This stage out will not be attempted:\n"
                amsg += '\t' + stageOutStr(stageOut) + '\n'
                amsg += str(ex)
                logging.exception(amsg)
                continue

        # no Rucio file catalog is initialized
        if not self.stageOuts_rfcs:
            raise StageOutInitError("===>Can not initialize Rucio file catalog")

        logging.info("==== Stageout configuration finish ====")

        return

    def initialiseOverride(self):
        """
        _initialiseOverride_

        Extract and verify that the Override parameters are all present

        """
        overrideConf = {
            "command": None,
            "option": None,
            "phedex-node": None,
            "lfn-prefix": None,
            }
        try:
            overrideConf['command'] = self.overrideConf['command']
            overrideConf['phedex-node'] = self.overrideConf['phedex-node']
            overrideConf['lfn-prefix'] = self.overrideConf['lfn-prefix']
        except Exception as ex:
            msg = "Unable to extract override parameters from config:\n"
            msg += str(ex)
            raise StageOutInitError(msg)
        if self.overrideConf.get('option') is not None:
            if len(self.overrideConf['option']) > 0:
                overrideConf['option'] = self.overrideConf['option']
            else:
                overrideConf['option'] = ""

        self.overrideConf = overrideConf

        msg = "=======StageOut Override Initialised:================\n"
        for key, val in viewitems(self.overrideConf):
            msg += " %s : %s\n" % (key, val)
        msg += "=====================================================\n"
        logging.info(msg)
        return

    def __call__(self, fileToStage):
        """
        _operator()_

        Use call to invoke transfers

        """
        lastException = Exception("empty exception")

        logging.info("==>Working on file: %s", fileToStage['LFN'])
        lfn = fileToStage['LFN']

        fileToStage['StageOutReport'] = []

        # // No override => use stage-out from site conf
        if not self.override:
            logging.info("===> Attempting %s Stage Outs", len(self.stageOuts))
            for stageOut_rfc in self.stageOuts_rfcs:
                try:
                    pfn = self.stageOut(lfn, fileToStage['PFN'], fileToStage.get('Checksums'), stageOut_rfc)
                    fileToStage['PFN'] = pfn
                    fileToStage['PNN'] = stageOut_rfc[0]['phedex-node']
                    fileToStage['StageOutCommand'] = stageOut_rfc[0]['command']
                    logging.info("attempting stageOut")
                    self.completedFiles[fileToStage['LFN']] = fileToStage
                    if lfn in self.failed:
                        del self.failed[lfn]

                    logging.info("===> Stage Out Successful: %s", fileToStage)
                    fileToStage = stageoutPolicyReport(fileToStage, None, None, 'LOCAL', 0)
                    return fileToStage
                except WMException as ex:
                    lastException = ex
                    logging.info("===> Stage Out Failure for file:")
                    logging.info("======>  %s\n", fileToStage['LFN'])
                    logging.info("======>  %s\n using this stage out", stageOutStr(stageOut_rfc[0]))
                    fileToStage = stageoutPolicyReport(fileToStage, stageOut_rfc[0]['phedex-node'],
                                                       stageOut_rfc[0]['command'], 'LOCAL', 60311)
                    continue
                except Exception as ex:
                    lastException = StageOutFailure("Error during local stage out",
                                                    error=str(ex))
                    logging.info("===> Stage Out Failure for file:\n")
                    logging.info("======>  %s\n", fileToStage['LFN'])
                    logging.info("======>  %s\n using this stage out", stageOutStr(stageOut_rfc[0]))
                    fileToStage = stageoutPolicyReport(fileToStage, stageOut_rfc[0]['phedex-node'],
                                                       stageOut_rfc[0]['command'], 'LOCAL', 60311)
                    continue

        else:
            logging.info("===> Attempting stage outs from override")
            try:
                pfn = self.stageOut(lfn, fileToStage['PFN'], fileToStage.get('Checksums'))
                fileToStage['PFN'] = pfn
                fileToStage['PNN'] = self.overrideConf['phedex-node']
                fileToStage['StageOutCommand'] = self.overrideConf['command']
                logging.info("attempting override stage out")
                self.completedFiles[fileToStage['LFN']] = fileToStage
                if lfn in self.failed:
                    del self.failed[lfn]

                logging.info("===> Stage Out Successful: %s", fileToStage)
                fileToStage = stageoutPolicyReport(fileToStage, None, None, 'OVERRIDE', 0)
                return fileToStage
            except Exception as ex:
                fileToStage = stageoutPolicyReport(fileToStage, self.overrideConf['phedex-node'], \
                                                   self.overrideConf['command'], 'OVERRIDE', 60310)
                lastException = ex

        raise lastException

    def stageOut(self, lfn, localPfn, checksums, stageOut_rfc=None):
        """
        Given the lfn and a pair of stage out and corresponding Rucio file catalog, stageOut_rfc, or override configuration invoke the stage out
        If use override configuration self.overrideConf should contain:
        command - the stage out impl plugin name to be used
        option - the option values to be passed to that command (None is allowed)
        lfn-prefix - the LFN prefix to generate the PFN
        phedex-node - the Name of the PNN to which the file is being xferred
        :param lfn: logical file name 
        :param localPfn: physical file name of file at local location (source) that will be staged out to another location (destination)
        :param checksums: check sum of file
        :param stageOut_rfc: a pair of stage out and corresponding Rucio file catalog
        """

        if not self.override:
            if not stageOut_rfc:
                msg = "Can not perform stage out for this lfn because of missing stage out information (stageOut_rfc is None or empty): \n %s" % lfn
                raise StageOutFailure(msg, LFN=lfn)
            command = stageOut_rfc[0]['command']
            options = stageOut_rfc[0]['option']
            pfn = searchRFC(stageOut_rfc[1], lfn)
            protocol = stageOut_rfc[1].preferredProtocol
            if pfn == None:
                msg = "Unable to match lfn to pfn: \n  %s" % lfn
                raise StageOutFailure(msg, LFN=lfn, StageOut=stageOutStr(stageOut_rfc[0]))
            try:
                impl = retrieveStageOutImpl(command)
            except Exception as ex:
                msg = "Unable to retrieve impl for local stage out:\n"
                msg += "Error retrieving StageOutImpl for command named: %s\n" % (
                    command,)
                raise StageOutFailure(msg, Command=command,
                                      LFN=lfn, ExceptionDetail=str(ex))
            impl.numRetries = self.numberOfRetries
            impl.retryPause = self.retryPauseTime

            try:
                impl(protocol, localPfn, pfn, options, checksums)
            except Exception as ex:
                msg = "Failure for stage out:\n"
                msg += str(ex)
                try:
                    import traceback
                    msg += traceback.format_exc()
                except AttributeError:
                    msg += "Traceback unavailable\n"
                raise StageOutFailure(msg, Command=command, Protocol=protocol,
                                      LFN=lfn, InputPFN=localPfn, TargetPFN=pfn)
            return pfn

        else:
            if self.overrideConf['lfn-prefix'] is None:
                msg = "Unable to match lfn to pfn during stage out because override lfn-prefix is None: \n %s" % lfn
                raise StageOutFailure(msg, LFN=lfn)

            pfn = "%s%s" % (self.overrideConf['lfn-prefix'], lfn)

            try:
                impl = retrieveStageOutImpl(self.overrideConf['command'])
            except Exception as ex:
                msg = "Unable to retrieve impl for override stage out:\n"
                msg += "Error retrieving StageOutImpl for command named: "
                msg += "%s\n" % self.overrideConf['command']
                raise StageOutFailure(msg, Command=self.overrideConf['command'],
                                      LFN=lfn, ExceptionDetail=str(ex))

            impl.numRetries = self.numberOfRetries
            impl.retryPause = self.retryPauseTime

            try:
                impl(self.overrideConf['command'], localPfn, pfn, self.overrideConf["option"], checksums)
            except Exception as ex:
                msg = "Failure for override stage out:\n"
                msg += str(ex)
                raise StageOutFailure(msg, Command=self.overrideConf['command'],
                                      LFN=lfn, InputPFN=localPfn, TargetPFN=pfn)

            return pfn

    def cleanSuccessfulStageOuts(self):
        """
        In the event of a failed stage out, this method can be called to cleanup the
        files that may have previously been staged out so that the job ends in a clear state
        of failure, rather than a partial success
        """

        for lfn, fileInfo in viewitems(self.completedFiles):
            pfn = fileInfo['PFN']
            command = fileInfo['StageOutCommand']
            msg = "Cleaning out file: %s\n" % lfn
            msg += "Removing PFN: %s" % pfn
            msg += "Using command implementation: %s\n" % command
            logging.info(msg)
            delManager = DeleteMgr(**self.overrideConf)
            try:
                delManager.deletePFN(pfn, lfn, command)
            except StageOutFailure as ex:
                msg = "Failed to cleanup staged out file after error:"
                msg += " %s\n%s" % (lfn, str(ex))
                logging.error(msg)



#!/usr/bin/env python
"""
_Save_

Oracle implementation of Masks.Save
"""

__all__ = []

from WMCore.WMBS.MySQL.Masks.Save import Save as SaveMasksMySQL


class Save(SaveMasksMySQL):
    sql = """INSERT INTO wmbs_job_mask (job, firstevent, lastevent, firstrun, lastrun, firstlumi, lastlumi, inclusivemask)
             SELECT :jobid, :firstevent, :lastevent, :firstrun, :lastrun, :firstlumi, :lastlumi, :inclusivemask FROM dual
             WHERE NOT EXISTS (SELECT * FROM wmbs_job_mask wjm2 WHERE
                               job = :jobid
                               AND wjm2.firstevent = :firstevent
                               AND wjm2.lastevent = :lastevent
                               AND wjm2.firstlumi = :firstlumi
                               AND wjm2.lastlumi = :lastlumi
                               AND wjm2.firstrun = :firstrun
                               AND wjm2.lastrun = :lastrun
                               AND wjm2.inclusivemask = :inclusivemask)"""

    def execute(self, jobid, mask, conn=None, transaction=False):
        if isinstance(mask, list):
            # Bulk commit
            # Hope you didn't send us a list of empty masks
            binds = []
            for m in mask:
                binds.append({"jobid": m['jobID'], 'firstevent': m['FirstEvent'], 'lastevent': m['LastEvent'],
                              'firstrun': m['FirstRun'], 'lastrun': m['LastRun'], 'firstlumi': m['FirstLumi'],
                              'lastlumi': m['LastLumi'], 'inclusivemask': m['inclusivemask']})
        else:
            # Simple one-part mask
            binds = {"jobid": jobid, 'firstevent': mask['FirstEvent'], 'lastevent': mask['LastEvent'],
                     'firstrun': mask['FirstRun'], 'lastrun': mask['LastRun'], 'firstlumi': mask['FirstLumi'],
                     'lastlumi': mask['LastLumi'], 'inclusivemask': mask['inclusivemask']}

            fail = True
            for key in binds:
                if key != 'jobid' and key != 'inclusivemask' and binds[key] != None:
                    # At least one of the keys contains something
                    fail = False
            if fail:
                return

        # Actually run the code
        self.dbi.processData(self.sql, binds, conn=conn, transaction=transaction)

        return



#!/usr/bin/env python
"""
_BulkNewReturn_

Oracle implementation of Fileset.BulkNewReturn
"""

__all__ = []



from WMCore.WMBS.MySQL.Fileset.BulkNewReturn import BulkNewReturn as MySQLBulkNewReturn

class BulkNewReturn(MySQLBulkNewReturn):
    """
    Does a bulk commit of Fileset, followed by returning their IDs

    """

    sql = """INSERT INTO wmbs_fileset (name, last_update, open)
             VALUES (:NAME, :LAST_UPDATE, :OPEN)"""



#!/usr/bin/env

"""
_MathAlgos_

Simple mathematical tools and tricks that might prove to
be useful.
"""
from __future__ import print_function, division
from builtins import str, range

import math
import decimal
import logging

from WMCore.WMException import WMException


class MathAlgoException(WMException):
    """
    Some simple math algo exceptions

    """
    pass


def getAverageStdDev(numList):
    """
    _getAverageStdDev_

    Given a list, calculate both the average and the
    standard deviation.
    """

    total   = 0.0
    average = 0.0
    stdBase = 0.0

    # Assemble the average
    skipped = 0
    for value in numList:
        try:
            if math.isnan(value) or math.isinf(value):
                skipped += 1
                continue
            else:
                total += value
        except TypeError:
            msg =  "Attempted to take average of non-numerical values.\n"
            msg += "Expected int or float, got %s: %s" % (value.__class__, value)
            logging.error(msg)
            logging.debug("FullList: %s", numList)
            raise MathAlgoException(msg)

    length = len(numList) - skipped
    if length < 1:
        return average, total

    average = total / length

    for value in numList:
        tmpValue = value - average
        stdBase += (tmpValue * tmpValue)

    stdDev = math.sqrt(stdBase / length)

    if math.isnan(average) or math.isinf(average):
        average = 0.0
    if math.isnan(stdDev) or math.isinf(average) or not decimal.Decimal(str(stdDev)).is_finite():
        stdDev = 0.0
    if not isinstance(stdDev, (int, float)):
        stdDev = 0.0

    return average, stdDev


def createHistogram(numList, nBins, limit):
    """
    _createHistogram_

    Create a histogram proxy (a list of bins) for a
    given list of numbers
    """

    average, stdDev = getAverageStdDev(numList = numList)

    underflow  = []
    overflow   = []
    histEvents = []
    histogram  = []
    for value in numList:
        if math.fabs(average - value) <= limit * stdDev:
            # Then we counted this event
            histEvents.append(value)
        elif average < value:
            overflow.append(value)
        elif average > value:
            underflow.append(value)

    if len(underflow) > 0:
        binAvg, binStdDev = getAverageStdDev(numList=underflow)
        histogram.append({'type': 'underflow',
                          'average': binAvg,
                          'stdDev': binStdDev,
                          'nEvents': len(underflow)})
    if len(overflow) > 0:
        binAvg, binStdDev = getAverageStdDev(numList=overflow)
        histogram.append({'type': 'overflow',
                          'average': binAvg,
                          'stdDev': binStdDev,
                          'nEvents': len(overflow)})
    if len(histEvents) < 1:
        # Nothing to do?
        return histogram

    histEvents.sort()
    upperBound = max(histEvents)
    lowerBound = min(histEvents)
    if lowerBound == upperBound:
        # This is a problem
        logging.debug("Only one value in the histogram!")
        nBins = 1
        upperBound = upperBound + 1
        lowerBound = lowerBound - 1
    binSize = (upperBound - lowerBound)/nBins
    binSize = floorTruncate(binSize)

    for x in range(nBins):
        lowerEdge = floorTruncate(lowerBound + (x * binSize))
        histogram.append({'type': 'standard',
                          'lowerEdge': lowerEdge,
                          'upperEdge': lowerEdge + binSize,
                          'average': 0.0,
                          'stdDev': 0.0,
                          'nEvents': 0})

    for bin_ in histogram:
        if bin_['type'] != 'standard':
            continue
        binList = []
        for value in histEvents:
            if value >= bin_['lowerEdge'] and value <= bin_['upperEdge']:
                # Then we're in the bin
                binList.append(value)
            elif value > bin_['upperEdge']:
                # Because this is a sorted list we are now out of the bin range
                # Calculate our values and break
                break
            else:
                continue

        # If we get here, it's because we're out of values in the bin
        # Time to do some math
        if len(binList) < 1:
            # Nothing to do here, leave defaults
            continue

        binAvg, binStdDev = getAverageStdDev(numList=binList)
        bin_['average'] = binAvg
        bin_['stdDev']  = binStdDev
        bin_['nEvents'] = len(binList)

    return histogram


def floorTruncate(value, precision=3):
    """
    _floorTruncate_

    Truncate a value to a set number of decimal points

    Always truncates to a LOWER value, this is so that using it for
    histogram binning creates values beneath the histogram lower edge.
    """
    prec = math.pow(10, precision)

    return math.floor(value * prec)/prec


def sortDictionaryListByKey(dictList, key, reverse=False):
    """
    _sortDictionaryListByKey_

    Given a list of dictionaries and a key with a numerical
    value, sort that dictionary in order of that key's value.

    NOTE: If the key does not exist, this will not raise an exception
    This is because this is used for sorting of performance histograms
    And not all histograms have the same value
    """

    return sorted(dictList, key=lambda k: float(k.get(key, 0.0)), reverse=reverse)


def getLargestValues(dictList, key, n=1):
    """
    _getLargestValues_

    Take a list of dictionaries, sort them by the value of a
    particular key, and return the n largest entries.

    Key must be a numerical key.
    """

    sortedList = sortDictionaryListByKey(dictList=dictList, key=key, reverse=True)

    return sortedList[:n]


def validateNumericInput(value):
    """
    _validateNumericInput_

    Check that the value is actually an usable number
    """
    value = float(value)
    try:
        if math.isnan(value) or math.isinf(value):
            return False
    except TypeError:
        return False

    return True


def calculateRunningAverageAndQValue(newPoint, n, oldM, oldQ):
    """
    _calculateRunningAverageAndQValue_

    Use the algorithm described in:
    Donald E. Knuth (1998). The Art of Computer Programming, volume 2: Seminumerical Algorithms, 3rd ed.., p. 232. Boston: Addison-Wesley.

    To calculate an average and standard deviation while getting data, the standard deviation
    can be obtained from the so-called Q value with the following equation:

    sigma = sqrt(Q/n)

    This is also contained in the function calculateStdDevFromQ in this module. The average is equal to M.
    """

    if not validateNumericInput(newPoint): raise MathAlgoException("Provided a non-valid newPoint")
    if not validateNumericInput(n): raise MathAlgoException("Provided a non-valid n")

    if n == 1:
        M = newPoint
        Q = 0.0
    else:
        if not validateNumericInput(oldM): raise MathAlgoException("Provided a non-valid oldM")
        if not validateNumericInput(oldQ): raise MathAlgoException("Provided a non-valid oldQ")
        M = oldM + (newPoint - oldM) / n
        Q = oldQ + ((n - 1) * (newPoint - oldM) * (newPoint - oldM) / n)

    return M, Q


def calculateStdDevFromQ(Q, n):
    """
    _calculateStdDevFromQ_

    If Q is the sum of the squared differences of some points to their average,
    then the standard deviation is given by:

    sigma = sqrt(Q/n)

    This function calculates that formula
    """
    if not validateNumericInput(Q): raise MathAlgoException("Provided a non-valid Q")
    if not validateNumericInput(n): raise MathAlgoException("Provided a non-valid n")

    sigma = math.sqrt(Q / n)

    if not validateNumericInput(sigma): return 0.0

    return sigma



#!/usr/bin/env python
"""
_GetCompletedWorkflows_

Oracle implementation of DBS3Buffer.GetCompletedWorkflows
"""




from WMComponent.DBS3Buffer.MySQL.GetCompletedWorkflows \
    import GetCompletedWorkflows as MySQLGetCompletedWorkflows

class GetCompletedWorkflows(MySQLGetCompletedWorkflows):
    """
    Oracle version
    """



#!/usr/bin/env python
"""
_New_

MySQL implementation of Fileset.New
"""

__all__ = []



import time

from WMCore.Database.DBFormatter import DBFormatter

class New(DBFormatter):
    sql = """INSERT INTO wmbs_fileset (name, last_update, open)
               VALUES (:NAME, :LAST_UPDATE, :OPEN)"""

    def getBinds(self, name = None, open = False):
        bindVars = {}
        bindVars["NAME"] = name
        bindVars["OPEN"] = int(open)
        bindVars["LAST_UPDATE"] = int(time.time())
        return bindVars

    def execute(self, name = None, open = False, conn = None,
                transaction = False):
        self.dbi.processData(self.sql, self.getBinds(name, open),
                         conn = conn, transaction = transaction)
        return



#!/usr/bin/env python
"""
_Credential_
Parent of all credentials class.
Childs class should implement methods of this class.
"""
from builtins import object
class Credential(object):
    """
    An abstract credential
    """

    def __init__( self, args ):

        self.serverKey = args.get( "server_key", "$HOME/.globus/hostkey.pem")
        self.serverCert = args.get( "server_cert", "$HOME/.globus/hostcert.pem")
        self.minTimeLeft = args.get( "min_time_left", 3600 )
        self.credServerPath = args.get( "cred_storage_path", '/tmp')
        self.userName = args.get( "userName", '')

    def create( self ):
        """
        Init the user credential
        """
        raise NotImplementedError

    def renew( self ):
        """
        Renew the user credential
        """
        raise NotImplementedError

    def get( self ):
        """
        Get user credential
        """
        raise NotImplementedError

    def store( self ):
        """
        Store user credential path in DB
        """
        raise NotImplementedError

    def check( self, credential=None ):
        """
        Check user credential
        """
        # Get the credential time left in second
        timeLeftLocal = self.getTimeLeft( credential )

        if timeLeftLocal and int( timeLeftLocal ) < self.minTimeLeft :
            return False
        else:
            return True

    def delegate( self, credential=None, serverRenewer = None ):
        """
        Delegate the user credential
        """
        raise NotImplementedError

    def getUserName( self, credential=None ):
        """
        Get the user name
        """
        raise NotImplementedError

    def getTimeLeft( self, credential=None ):
        """
        Get credential time left
        """
        raise NotImplementedError

    def destroy( self, credential = None ):
        """
        Destroy user credential
        """
        raise NotImplementedError



#!/usr/bin/env python

"""
Oracle implementation of AddChecksumByLFN
"""





from WMCore.WMBS.MySQL.Files.AddChecksumByLFN import AddChecksumByLFN as MySQLAddChecksumByLFN

class AddChecksumByLFN(MySQLAddChecksumByLFN):
    """
    Add Checksums using LFN as file identifier

    """
    sql = """INSERT INTO wmbs_file_checksums (fileid, typeid, cksum)
             SELECT (SELECT id FROM wmbs_file_details WHERE lfn = :lfn),
              (SELECT id FROM wmbs_checksum_type WHERE type = :cktype),
              :cksum FROM dual
             WHERE NOT EXISTS (SELECT fileid FROM wmbs_file_checksums WHERE
                               fileid = (SELECT id FROM wmbs_file_details WHERE lfn = :lfn)
                               AND typeid = (SELECT id FROM wmbs_checksum_type WHERE type = :cktype))"""



#!/usr/bin/env python
"""
Default function for Job Maker
"""
from __future__ import print_function
__all__ = []




from WMCore.Agent.BaseHandler import BaseHandler
from WMCore.ThreadPool.ThreadPool import ThreadPool

import threading



class MakeJob(BaseHandler):
    """
    Default handler for Make Job Events.
    """

    def __init__(self, component):
        BaseHandler.__init__(self, component)
        # define a slave threadpool (this is optional
        # and depends on the developer deciding how he/she
        # wants to implement certain logic.

        self.nThreads = 5

        self.threadpool = ThreadPool(\
            "WMCore.WMSpec.Makers.Handlers.MakeJobSlave", \
            self.component, 'MakeJob', \
            self.nThreads)
        myThread = threading.currentThread()
        myThread.msgService.purgeMessages()

     # this we overload from the base handler
    def __call__(self, event, payload):
        """
        Handles the event with payload, by sending it to the threadpool.
        """
        # as we defined a threadpool we can enqueue our item
        # and move to the next.
        self.threadpool.enqueue(event, payload)
        print("Thread with payload " + str(payload) + " is enqueued")



#!/usr/bin/env python
"""
_GetFinishedWorkflows_

MySQL implementation of Workflows.GetFinishedWorkflows

Created on Aug 29, 2012

@author: dballest
"""

from WMCore.Database.DBFormatter import DBFormatter

class GetFinishedWorkflows(DBFormatter):
    """
    A workflow is finished when all subscriptions associated to it are finished
    It assumes that a spec is not shared by two workflows of the same name.
    """

    incompleteWf = """ SELECT wmbs_workflow.name
                                  FROM wmbs_workflow
                                      INNER JOIN wmbs_subscription ON
                                         wmbs_subscription.workflow = wmbs_workflow.id
                                      INNER JOIN wmbs_sub_types ON
                                         wmbs_sub_types.id = wmbs_subscription.subtype
                                  WHERE wmbs_subscription.finished = 0
                                       AND wmbs_sub_types.name %(include)s IN ('LogCollect', 'Cleanup')
                                  GROUP BY wmbs_workflow.name """

    sql = """ SELECT wmbs_workflow.name, wmbs_workflow.spec,
                        wmbs_workflow.id AS workflow_id, wmbs_subscription.id AS sub_id
                 FROM wmbs_subscription
                    INNER JOIN wmbs_workflow ON
                         wmbs_workflow.id = wmbs_subscription.workflow
                    INNER JOIN wmbs_sub_types ON
                         wmbs_sub_types.id = wmbs_subscription.subtype
                    WHERE wmbs_sub_types.name %(include)s IN ('LogCollect', 'Cleanup')  AND
                            wmbs_workflow.name NOT IN (""" + incompleteWf + """ )"""

    def execute(self, onlySecondary=False, conn=None, transaction=False):
        """
        _execute_

        onlySecondary if set it True gets the complete subscription for the only LogCollect and Cleanup type.
        if False, gets the finished subscription excluding LogCollect and Cleanup tasks
        This DAO is a nested dictionary with the following structure:
        {<workflowName? : {spec : <specURL>,
                           workflows : {<workflowID> : [<subId1>, <subId2>]}
                          }
        }
        """
        if not onlySecondary:
            include = {'include': 'NOT' }
        else:
            include = {'include': '' }

        sql = self.sql % include

        #Get the completed workflows and subscriptions
        result = self.dbi.processData(sql, conn=conn, transaction=transaction)
        wfsAndSubs = self.formatDict(result)

        wfs = {}
        for entry in wfsAndSubs:
            if entry['name'] not in wfs:
                wfs[entry['name']] = {'workflows' : {}}
            wfs[entry['name']]['spec'] = entry['spec']
            if entry['workflow_id'] not in wfs[entry['name']]['workflows']:
                wfs[entry['name']]['workflows'][entry['workflow_id']] = []
            wfs[entry['name']]['workflows'][entry['workflow_id']].append(entry['sub_id'])

        return wfs



#!/usr/bin/env python
"""
_LoadByStatus_

Oracle implementation for loading a job by scheduler status
"""


from WMCore.BossAir.MySQL.LoadByStatus import LoadByStatus as MySQLLoadByStatus

class LoadByStatus(MySQLLoadByStatus):
    """
    _LoadByStatus_

    Load all jobs with a certain scheduler status
    """



#!/usr/bin/env python
"""
_New_

Oracle implementation of JobGroup.New
"""

__all__ = []



import time

from WMCore.WMBS.MySQL.JobGroup.New import New as NewJobGroupMySQL

class New(NewJobGroupMySQL):
    sql = """INSERT INTO wmbs_jobgroup (subscription, guid, output, last_update)
             VALUES (:subscription, :guid, :output, %d)""" % time.time()
